import numpy as np
import pytest

from nengo.exceptions import ValidationError
from nengo.utils.least_squares_solvers import (
    SVD,
    BlockConjgrad,
    Cholesky,
    Conjgrad,
    ConjgradScipy,
    LSMRScipy,
    RandomizedSVD,
)


def run_solver(solver, m=30, n=20, d=1, cond=10, sys_rng=np.random, **kwargs):
    A = sys_rng.uniform(-1, 1, size=(m, n))
    x = sys_rng.uniform(-1, 1, size=(n, d))

    # set condition number of system
    assert cond > 1
    U, s, VH = np.linalg.svd(A, full_matrices=False)
    smax, smin = s.max(), s.min()
    new_min = smax / cond
    s = (s - smin) / (smax - smin) * (smax - new_min) + new_min
    A = np.dot(U * s, VH)

    # solve system
    y = np.dot(A, x)
    x2, _ = solver(A, y, **kwargs)

    return x, x2, A


@pytest.mark.parametrize("cond, sigma", [(5, 1e-4), (500, 1e-6), (1e5, 1e-8)])
def test_cholesky(cond, sigma, rng, allclose):
    rng = np.random
    solver = Cholesky()
    x, x2, _ = run_solver(solver, d=2, cond=cond, sys_rng=rng, sigma=sigma)

    tol = np.sqrt(cond) * 1e-7  # this is a guesstimate, and also depends on sigma
    assert allclose(x2, x, atol=tol, rtol=tol)


@pytest.mark.parametrize("cond, sigma, tol", [(5, 1e-4, 1e-2), (50, 1e-5, 1e-4)])
def test_conjgrad_scipy(cond, sigma, tol, rng, allclose):
    pytest.importorskip("scipy")
    solver = ConjgradScipy(tol=tol)
    x, x2, _ = run_solver(solver, d=2, cond=cond, sys_rng=rng, sigma=sigma)

    tol = cond * tol
    assert allclose(x2, x, atol=tol, rtol=tol)


@pytest.mark.parametrize("cond, sigma, tol", [(5, 1e-3, 1e-4), (10, 1e-3, 1e-4)])
def test_lsmr_scipy(cond, sigma, tol, rng, allclose):
    pytest.importorskip("scipy")
    solver = LSMRScipy(tol=tol)
    x, x2, _ = run_solver(solver, d=2, cond=cond, sys_rng=rng, sigma=sigma)

    tol = cond * tol
    assert allclose(x2, x, atol=tol, rtol=tol)


@pytest.mark.parametrize(
    "cond, sigma, tol, maxiters",
    [
        (5, 1e-8, 1e-2, None),  # standard run
        (50, 1e-8, 1e-4, 100),  # precision run (extra iterations help convergence)
        (1.1, 1e-15, 0, 1000),  # hit "no perceptible change in p" line
    ],
)
def test_conjgrad(cond, sigma, tol, maxiters, rng, allclose):
    solver = Conjgrad(tol=tol, maxiters=maxiters)
    x, x2, A = run_solver(solver, d=2, cond=cond, sys_rng=rng, sigma=sigma)

    # conjgrad stopping tol is based on residual in y, so compare y (`y = A.dot(x)`)
    # (this particularly helps in the ill-conditioned case)
    tol = cond * max(tol, 1e-5)
    assert allclose(A.dot(x2), A.dot(x), atol=tol, rtol=tol)


def test_conjgrad_errors():
    solver = Conjgrad(X0=np.ones((1, 1)))
    with pytest.raises(ValidationError, match=r"Must be shape \(2, 2\)"):
        solver(A=np.ones((2, 2)), Y=np.ones((2, 2)), sigma=0.001)


@pytest.mark.parametrize("cond, tol", [(5, 1e-2), (100, 1e-3)])
def test_blockconjgrad(cond, tol, rng, allclose):
    x, x2, _ = run_solver(
        BlockConjgrad(tol=tol), d=5, cond=cond, sys_rng=rng, sigma=1e-8
    )
    assert allclose(x2, x, atol=tol, rtol=tol)


def test_blockconjgrad_errors():
    solver = BlockConjgrad(X0=np.ones((1, 1)))
    with pytest.raises(ValidationError, match=r"Must be shape \(2, 2\)"):
        solver(A=np.ones((2, 2)), Y=np.ones((2, 2)), sigma=0.1)


@pytest.mark.parametrize("cond", [5, 1000])
def test_svd(cond, rng, allclose):
    x, x2, _ = run_solver(SVD(), d=5, cond=cond, sys_rng=rng, sigma=1e-8)
    assert allclose(x2, x, atol=1e-8, rtol=1e-8)


@pytest.mark.parametrize("cond", [5, 1000])
def test_randomized_svd_fallback(cond, rng, allclose):
    """Test the specific case where RandomizedSVD falls back to SVD"""
    pytest.importorskip("sklearn")
    m, n = 30, 20
    solver = RandomizedSVD(n_components=min(m, n))
    x, x2, _ = run_solver(solver, m=m, n=n, d=5, cond=cond, sys_rng=rng, sigma=1e-8)
    assert allclose(x2, x, atol=1e-8, rtol=1e-8)



from pytest_allclose import report_rmses

from nengo.rc import rc


def pytest_runtest_setup(item):
    rc.reload_rc([])
    rc["decoder_cache"]["enabled"] = "False"
    rc["exceptions"]["simplified"] = "False"
    rc["nengo.Simulator"]["fail_fast"] = "True"
    rc["progress"]["progress_bar"] = "False"


def pytest_terminal_summary(terminalreporter):
    report_rmses(terminalreporter)



import nengo
from nengo.exceptions import ValidationError
from nengo.spa.module import Module


class Bind(Module):
    """A module for binding together two inputs.

    Binding is done with circular convolution. For more details on how
    this is computed, see the underlying `~.networks.CircularConvolution`
    network.

    Parameters
    ----------
    dimensions : int
        Number of dimensions for the two vectors to be compared.
    vocab : Vocabulary, optional
        The vocabulary to use to interpret the vectors. If None,
        the default vocabulary for the given dimensionality is used.
    n_neurons : int, optional
        Number of neurons to use in each product computation.
    invert_a, invert_b : bool, optional
        Whether to reverse the order of elements in either
        the first input (``invert_a``) or the second input (``invert_b``).
        Flipping the second input will make the network perform circular
        correlation instead of circular convolution.
    input_magnitude : float, optional
        The expected magnitude of the vectors to be convolved.
        This value is used to determine the radius of the ensembles
        computing the element-wise product.

    label : str, optional
        A name for the ensemble. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.
    add_to_container : bool, optional
        Determines if this Network will be added to the current container.
        If None, will be true if currently within a Network.
    """

    def __init__(
        self,
        dimensions,
        vocab=None,
        n_neurons=200,
        invert_a=False,
        invert_b=False,
        input_magnitude=1.0,
        label=None,
        seed=None,
        add_to_container=None,
    ):
        super().__init__(label, seed, add_to_container)
        if vocab is None:
            # use the default vocab for this number of dimensions
            vocab = dimensions
        elif vocab.dimensions != dimensions:
            raise ValidationError(
                "Dimensionality of given vocabulary (%d) does "
                "not match dimensionality of buffer (%d)"
                % (vocab.dimensions, dimensions),
                attr="dimensions",
                obj=self,
            )

        with self:
            self.cc = nengo.networks.CircularConvolution(
                n_neurons,
                dimensions,
                invert_a,
                invert_b,
                input_magnitude=input_magnitude,
            )
            self.A = self.cc.input_a
            self.B = self.cc.input_b
            self.output = self.cc.output

        self.inputs = dict(A=(self.A, vocab), B=(self.B, vocab))
        self.outputs = dict(default=(self.output, vocab))



import inspect
import sys

from nengo.utils.magic import BoundFunctionWrapper, ObjectProxy, decorator


class RunState:
    state = None  # Used to make sure decorators are running


def _test_decorated(obj):

    # Make sure decorator works
    RunState.state = "not run"
    obj(0, 0)
    assert RunState.state == "run"
    RunState.state = "not run"

    # Make sure decorated function looks like non-decorated
    assert obj.__name__ == "f"
    assert obj.__doc__ == "Return 1."


def test_function():
    """Test that @decorator works on plain functions."""

    @decorator
    def test_decorator(wrapped, instance, args, kwargs):
        RunState.state = "run"
        assert instance is None
        return wrapped(*args, **kwargs)

    @test_decorator
    def f(a, b):
        """Return 1."""
        return 1

    _test_decorated(f)

    # Make sure introspection works
    assert inspect.getfullargspec(f).args == ["a", "b"]
    assert inspect.getsource(f) == (
        "    @test_decorator\n"
        "    def f(a, b):\n"
        '        """Return 1."""\n'
        "        return 1\n"
    )


def test_boundfunction():
    """Test that @decorator works on bound functions."""

    @decorator
    def test_decorator(wrapped, instance, args, kwargs):
        RunState.state = "run"
        assert instance is not None
        assert type(instance).__name__ == "Test"
        return wrapped(*args, **kwargs)

    class Test:
        @test_decorator
        def f(self, a, b):
            """Return 1."""
            return 1

    inst = Test()
    _test_decorated(inst.f)

    # Make sure introspection works
    assert inspect.getfullargspec(inst.f).args == ["self", "a", "b"]
    assert inspect.getsource(inst.f) == (
        "        @test_decorator\n"
        "        def f(self, a, b):\n"
        '            """Return 1."""\n'
        "            return 1\n"
    )


def test_staticmethod():
    """Test that @decorator works on staticmethods."""

    @decorator
    def test_decorator(wrapped, instance, args, kwargs):
        RunState.state = "run"
        assert instance is None
        return wrapped(*args, **kwargs)

    # --- Decorator before staticmethod
    class TestBeforeStaticmethod:
        @test_decorator
        @staticmethod
        def f(a, b):
            """Return 1."""
            return 1

    _test_decorated(TestBeforeStaticmethod.f)
    inst = TestBeforeStaticmethod()
    _test_decorated(inst.f)

    # Make sure introspection works
    assert inspect.getfullargspec(inst.f).args == ["a", "b"]
    assert inspect.getsource(inst.f) == (
        "        @test_decorator\n"
        "        @staticmethod\n"
        "        def f(a, b):\n"
        '            """Return 1."""\n'
        "            return 1\n"
    )

    # --- Decorator after staticmethod
    class TestAfterStaticmethod:
        @staticmethod
        @test_decorator
        def f(a, b):
            """Return 1."""
            return 1

    _test_decorated(TestAfterStaticmethod.f)
    inst = TestAfterStaticmethod()
    _test_decorated(inst.f)

    # Make sure introspection works
    assert inspect.getfullargspec(inst.f).args == ["a", "b"]
    assert inspect.getsource(inst.f) == (
        "        @staticmethod\n"
        "        @test_decorator\n"
        "        def f(a, b):\n"
        '            """Return 1."""\n'
        "            return 1\n"
    )


def test_classmethod():
    """Test that @decorator works on classmethods."""

    @decorator
    def test_decorator(wrapped, instance, args, kwargs):
        RunState.state = "run"
        valid_names = {"TestBeforeStaticmethod", "TestAfterStaticmethod"}
        assert (
            instance is None
            and args[0].__name__ in valid_names
            or instance.__name__ in valid_names
        )
        return wrapped(*args, **kwargs)

    # --- Decorator before classmethod
    class TestBeforeStaticmethod:
        @test_decorator
        @classmethod
        def f(cls, a, b):
            """Return 1."""
            return 1

    _test_decorated(TestBeforeStaticmethod.f)
    inst = TestBeforeStaticmethod()
    _test_decorated(inst.f)

    # Make sure introspection works
    assert inspect.getfullargspec(inst.f).args == ["cls", "a", "b"]
    assert inspect.getsource(inst.f) == (
        "        @test_decorator\n"
        "        @classmethod\n"
        "        def f(cls, a, b):\n"
        '            """Return 1."""\n'
        "            return 1\n"
    )

    # --- Decorator after staticmethod
    class TestAfterStaticmethod:
        @classmethod
        @test_decorator
        def f(cls, a, b):
            """Return 1."""
            return 1

    _test_decorated(TestAfterStaticmethod.f)
    inst = TestAfterStaticmethod()
    _test_decorated(inst.f)

    # Make sure introspection works
    assert inspect.getfullargspec(inst.f).args == ["cls", "a", "b"]
    assert inspect.getsource(inst.f) == (
        "        @classmethod\n"
        "        @test_decorator\n"
        "        def f(cls, a, b):\n"
        '            """Return 1."""\n'
        "            return 1\n"
    )


def test_class():
    """Test that @decorator works on classes."""

    @decorator
    def test_decorator(wrapped, instance, args, kwargs):
        RunState.state = "run"
        assert instance is None
        inst = wrapped(*args, **kwargs)
        inst.ran = True
        return inst

    @test_decorator
    class f:
        """Return 1."""

        def __init__(self, a, b):
            self.a = a
            self.b = b

    _test_decorated(f)
    inst = f("a", "b")
    assert inst.a == "a" and inst.b == "b"
    assert inst.ran
    assert type(inst) == f.__wrapped__
    assert type(inst) == f.__wrapped__

    # Make sure introspection works
    source = inspect.getsource(f)

    target_source = (
        "    class f:\n"
        '        """Return 1."""\n'
        "\n"
        "        def __init__(self, a, b):\n"
        "            self.a = a\n"
        "            self.b = b\n"
    )

    if sys.version_info >= (3, 9, 0):
        # for Python>=3.9 the decorator is included in the source
        target_source = "    @test_decorator\n" + target_source

    assert source == target_source


def test_class_decorator():
    """Test that `decorator` works on a class method"""

    class TestA:
        @decorator
        @classmethod
        def test_decorator(cls, wrapped, instance, args, kwargs):
            RunState.state = "run"
            assert instance is not None
            assert type(instance).__name__ == "Test"
            return wrapped(*args, **kwargs)

    class Test:
        @TestA.test_decorator
        def f(self, a, b):
            """Return 1."""
            return 1

    inst = Test()
    _test_decorated(inst.f)

    # Make sure introspection works
    assert inspect.getfullargspec(inst.f).args == ["self", "a", "b"]
    assert inspect.getsource(inst.f) == (
        "        @TestA.test_decorator\n"
        "        def f(self, a, b):\n"
        '            """Return 1."""\n'
        "            return 1\n"
    )


def test_instance_decorator():
    """Test that `decorator` works on a class instance method"""

    class TestA:
        @decorator
        def test_decorator(self, wrapped, instance, args, kwargs):
            RunState.state = "run"
            assert instance is not None
            assert type(instance).__name__ == "Test"
            return wrapped(*args, **kwargs)

    test_a = TestA()

    class Test:
        @test_a.test_decorator
        def f(self, a, b):
            """Return 1."""
            return 1

    inst = Test()
    _test_decorated(inst.f)

    # Make sure introspection works
    assert inspect.getfullargspec(inst.f).args == ["self", "a", "b"]
    assert inspect.getsource(inst.f) == (
        "        @test_a.test_decorator\n"
        "        def f(self, a, b):\n"
        '            """Return 1."""\n'
        "            return 1\n"
    )


def test_objectproxy():
    """tests functions of ObjectProxy"""

    class Test:
        """my docstring"""

        __annotations__ = "testannotations"
        __name__ = "testname"

        testvar = "testval"

    obj = Test()
    proxy = ObjectProxy(obj)

    # --- test ObjectProxyMethods
    assert proxy.__dict__ is obj.__dict__
    assert proxy.__doc__ is obj.__doc__
    assert proxy.__module__ is obj.__module__

    # --- test ObjectProxy
    assert proxy.__annotations__ is obj.__annotations__
    assert proxy.__name__ is obj.__name__
    assert proxy.__class__ is obj.__class__

    assert set(dir(proxy)) == set(dir(obj))
    assert hash(proxy) == hash(obj)
    assert str(proxy) == str(obj)

    proxy.newattr = "newval"  # pylint: disable=assigning-non-slot
    assert obj.newattr == "newval"
    assert proxy.newattr == "newval"

    assert repr(proxy).startswith("<ObjectProxy at ")


def test_boundfunctionwrapper():
    """tests functions of BoundFunctionWrapper"""

    class MyParentHelper:
        def __get__(self, a, b=None):
            return 0

    class MyParent:
        __wrapped__ = MyParentHelper

    class MyWrapped:
        value = 1

    class MyFunction:
        def __init__(self, b, c, d, e):
            return None

        def __call__(self, b):
            return True

    function = MyFunction

    wrapped = MyWrapped
    parent = MyParent

    instance = BoundFunctionWrapper(wrapped, None, function, "function", parent)

    assert (
        str(BoundFunctionWrapper.__get__(instance, instance, BoundFunctionWrapper))
        == "0"
    )

    instance2 = BoundFunctionWrapper(wrapped, "Not None", 1, "function", parent)

    assert (
        str(BoundFunctionWrapper.__get__(instance2, instance2, BoundFunctionWrapper))
        == "<class 'nengo.utils.tests.test_magic"
        ".test_boundfunctionwrapper.<locals>.MyWrapped'>"
    )

    assert str(instance(1, 2)).startswith(
        "<nengo.utils.tests.test_magic."
        "test_boundfunctionwrapper.<locals>.MyFunction object at "
    )



import warnings

import numpy as np

from nengo.exceptions import SpaModuleError
from nengo.network import Network
from nengo.spa.module import Module
from nengo.spa.utils import enable_spa_params, similarity
from nengo.spa.vocab import Vocabulary


class SPA(Network):
    """Base class for SPA models.

    This expands the standard `.Network` system to support structured
    connections that use Semantic Pointers and associated vocabularies
    in their definitions.

    To build a SPA model, you can either use ``with`` or create a subclass
    of this SPA class.

    If you use the ``with`` statement, any attribute added to the SPA network
    will be accessible for SPA connections.

    If you chose to create a subclass, any `~.spa.module.Module` object that
    is assigned to a member variable will automatically be accessible by the
    SPA connection system.

    As an example, the following code will build three modules
    (two buffers and a memory) that can be referred to as ``a``, ``b``,
    and ``c``, respectively.

    First, the example with a ``with`` statement::

        example = spa.Spa()

        with example:
            example.a = spa.Buffer(dimensions=8)
            example.b = spa.Buffer(dimensions=16)
            example.c = spa.Memory(dimensions=8)

    Now, the example with a subclass::

        class Example(spa.SPA):
            def __init__(self):
                with self:
                    self.a = spa.Buffer(dimensions=8)
                    self.b = spa.Buffer(dimensions=16)
                    self.c = spa.Memory(dimensions=8)

    These names can be used by special modules that are aware of these
    names. As an example, the `.Cortical` module allows you to form connections
    between these modules in ways that are aware of semantic pointers::

        with example:
            example.a = spa.Buffer(dimensions=8)
            example.b = spa.Buffer(dimensions=16)
            example.c = spa.Memory(dimensions=8)
            example.cortical = spa.Cortical(spa.Actions(
                    'b=a*CAT', 'c=b*~CAT'))

    For complex cognitive control, the key modules are the `.spa.BasalGanglia`
    and the `.spa.Thalamus`. Together, these allow us to define complex actions
    using the `.spa.Actions` syntax::

        class SequenceExample(spa.SPA):
            def __init__(self):
                self.state = spa.Memory(dimensions=32)

                actions = spa.Actions('dot(state, A) --> state=B',
                                      'dot(state, B) --> state=C',
                                      'dot(state, C) --> state=D',
                                      'dot(state, D) --> state=E',
                                      'dot(state, E) --> state=A')

                self.bg = spa.BasalGanglia(actions=actions)
                self.thal = spa.Thalamus(self.bg)

    .. deprecated:: 3.0.0
    """

    def __init__(self, label=None, seed=None, add_to_container=None, vocabs=None):
        super().__init__(label, seed, add_to_container)

        warnings.warn(
            DeprecationWarning(
                "The nengo.spa module is deprecated. Please switch to using the "
                "NengoSPA project <https://www.nengo.ai/nengo-spa/>."
            )
        )

        vocabs = [] if vocabs is None else vocabs
        enable_spa_params(self)
        self._modules = {}
        self._default_vocabs = {}

        for vo in vocabs:
            if vo.dimensions in self._default_vocabs:
                warnings.warn(
                    "Duplicate vocabularies with dimension %d. "
                    "Using the last entry in the vocab list with "
                    "that dimensionality." % (vo.dimensions)
                )
            self._default_vocabs[vo.dimensions] = vo

        self._initialized = True

    def __setstate__(self, state):
        if "_initialized" in state:
            del state["_initialized"]
        super().__setstate__(state)
        setattr(self, "_initialized", True)

    def __setattr__(self, key, value):
        """A setattr that handles Modules being added specially.

        This is so that we can use the variable name for the Module as
        the name that all of the SPA system will use to access that module.
        """
        if not hasattr(self, "_initialized"):
            return super().__setattr__(key, value)

        if hasattr(self, key) and isinstance(getattr(self, key), Module):
            raise SpaModuleError(
                "Cannot re-assign module-attribute %s to %s. "
                "SPA module-attributes can only be assigned "
                "once." % (key, value)
            )
        super().__setattr__(key, value)
        if isinstance(value, Module):
            if value.label is None:
                value.label = key
            self._modules[key] = value
            for k, (obj, v) in value.inputs.items():
                if type(v) == int:
                    value.inputs[k] = (obj, self.get_default_vocab(v))
                self.config[obj].vocab = value.inputs[k][1]
            for k, (obj, v) in value.outputs.items():
                if type(v) == int:
                    value.outputs[k] = (obj, self.get_default_vocab(v))
                self.config[obj].vocab = value.outputs[k][1]

            value.on_add(self)

    def __exit__(self, ex_type, ex_value, traceback):
        super().__exit__(ex_type, ex_value, traceback)
        if ex_type is not None:
            # re-raise the exception that triggered this __exit__
            return False

        module_list = frozenset(self._modules.values())
        for net in self.networks:
            # Since there are no attributes to distinguish what's been added
            # and what hasn't, we have to ask the network
            if isinstance(net, Module) and (net not in module_list):
                raise SpaModuleError(
                    "%s must be set as an attribute of a SPA network" % (net)
                )

    def get_module(self, name):
        """Return the module for the given name."""
        if name in self._modules:
            return self._modules[name]
        elif "_" in name:
            module, name = name.rsplit("_", 1)
            if module in self._modules:
                return self._modules[module]
        raise SpaModuleError("Could not find module %r" % name)

    def get_default_vocab(self, dimensions):
        """Return a Vocabulary with the desired dimensions.

        This will create a new default Vocabulary if one doesn't exist.
        """

        # If seed is set, create rng based off that seed.
        # Otherwise, just use the default NumPy rng.
        rng = None if self.seed is None else np.random.RandomState(self.seed)

        if dimensions not in self._default_vocabs:
            self._default_vocabs[dimensions] = Vocabulary(dimensions, rng=rng)
        return self._default_vocabs[dimensions]

    def get_module_input(self, name):
        """Return the object to connect into for the given name.

        The name will be either the same as a module, or of the form
        ``<module_name>_<input_name>``.
        """
        if name in self._modules and "default" in self._modules[name].inputs:
            return self._modules[name].inputs["default"]
        elif "_" in name:
            module, name = name.rsplit("_", 1)
            if module in self._modules:
                m = self._modules[module]
                if name in m.inputs:
                    return m.inputs[name]
        raise SpaModuleError("Could not find module input %r" % name)

    def get_module_inputs(self):
        for name, module in self._modules.items():
            for input in module.inputs:
                if input == "default":
                    yield name
                else:
                    yield "%s_%s" % (name, input)

    def get_input_vocab(self, name):
        return self.get_module_input(name)[1]

    def get_module_output(self, name):
        """Return the object to connect into for the given name.

        The name will be either the same as a module, or of the form
        ``<module_name>_<output_name>``.
        """
        if name in self._modules:
            return self._modules[name].outputs["default"]
        elif "_" in name:
            module, name = name.rsplit("_", 1)
            if module in self._modules:
                m = self._modules[module]
                if name in m.outputs:
                    return m.outputs[name]
        raise SpaModuleError("Could not find module output %r" % name)

    def get_module_outputs(self):
        for name, module in self._modules.items():
            for output in module.outputs:
                if output == "default":
                    yield name
                else:
                    yield "%s_%s" % (name, output)

    def get_output_vocab(self, name):
        return self.get_module_output(name)[1]

    def similarity(self, data, probe, vocab=None):
        """Return the similarity between the probed data and ``vocab``.

        If no vocabulary is provided, the vocabulary associated with
        ``probe.target`` will be used.

        Parameters
        ----------
        data: SimulationData
            Collection of simulation data returned by sim.run() function call.
        probe: Probe
            Probe with desired data.
        vocab : Vocabulary, optional
            The vocabulary to compare with. If None, uses the vocabulary
            associated with ``probe.target``.
        """
        if vocab is None:
            vocab = self.config[probe.target].vocab
        return similarity(data[probe], vocab)



import numpy as np
import pytest

import nengo
from nengo import spa
from nengo.exceptions import SpaParseError


def thalamus_net(d=2, n=20, seed=None):
    model = spa.SPA(seed=seed)

    with model:
        model.vision = spa.Buffer(dimensions=d, neurons_per_dimension=n)
        model.vision2 = spa.Buffer(dimensions=d, neurons_per_dimension=n)
        model.motor = spa.Buffer(dimensions=d, neurons_per_dimension=n)
        model.motor2 = spa.Buffer(dimensions=d * 2, neurons_per_dimension=n)

        actions = spa.Actions(
            "dot(vision, A) --> motor=A, motor2=vision*vision2",
            "dot(vision, B) --> motor=vision, motor2=vision*A*~B",
            "dot(vision, ~A) --> motor=~vision, motor2=~vision*vision2",
        )
        model.bg = spa.BasalGanglia(actions)
        model.thalamus = spa.Thalamus(model.bg)

        def input_f(t):
            if t < 0.1:
                return "A"
            elif t < 0.3:
                return "B"
            elif t < 0.5:
                return "~A"
            else:
                return "0"

        model.input = spa.Input(vision=input_f, vision2="B*~A")

    return model


@pytest.mark.slow
def test_thalamus(Simulator, plt, seed):
    model = thalamus_net(d=16, n=80, seed=seed)

    with model:
        input, vocab = model.get_module_input("motor")
        input2, vocab2 = model.get_module_input("motor2")

        p = nengo.Probe(input, "output", synapse=0.03)
        p2 = nengo.Probe(input2, "output", synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)

    t = sim.trange()
    data = vocab.dot(sim.data[p].T)
    data2 = vocab2.dot(sim.data[p2].T)

    plt.subplot(2, 1, 1)
    plt.plot(t, data.T)
    plt.subplot(2, 1, 2)
    plt.plot(t, data2.T)

    # Action 1
    assert data[0, t == 0.1] > 0.8
    assert data[1, t == 0.1] < 0.2
    assert data2[0, t == 0.1] < 0.35
    assert data2[1, t == 0.1] > 0.4
    # Action 2
    assert data[0, t == 0.3] < 0.2
    assert data[1, t == 0.3] > 0.8
    assert data2[0, t == 0.3] > 0.5
    assert data2[1, t == 0.3] < 0.3
    # Action 3
    assert data[0, t == 0.5] > 0.8
    assert data[1, t == 0.5] < 0.2
    assert data2[0, t == 0.5] < 0.5
    assert data2[1, t == 0.5] > 0.4


def test_routing(Simulator, seed, plt):
    D = 3
    model = spa.SPA(seed=seed)
    with model:
        model.ctrl = spa.Buffer(16, label="ctrl")

        def input_func(t):
            if t < 0.2:
                return "A"
            elif t < 0.4:
                return "B"
            else:
                return "C"

        model.input = spa.Input(ctrl=input_func)

        model.buff1 = spa.Buffer(D, label="buff1")
        model.buff2 = spa.Buffer(D, label="buff2")
        model.buff3 = spa.Buffer(D, label="buff3")

        node1 = nengo.Node([0, 1, 0])
        node2 = nengo.Node([0, 0, 1])

        nengo.Connection(node1, model.buff1.state.input)
        nengo.Connection(node2, model.buff2.state.input)

        actions = spa.Actions(
            "dot(ctrl, A) --> buff3=buff1",
            "dot(ctrl, B) --> buff3=buff2",
            "dot(ctrl, C) --> buff3=buff1*buff2",
        )
        model.bg = spa.BasalGanglia(actions)
        model.thal = spa.Thalamus(model.bg)

        buff3_probe = nengo.Probe(model.buff3.state.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.6)

    data = sim.data[buff3_probe]

    plt.plot(sim.trange(), data)

    valueA = np.mean(data[150:200], axis=0)  # should be [0, 1, 0]
    valueB = np.mean(data[350:400], axis=0)  # should be [0, 0, 1]
    valueC = np.mean(data[550:600], axis=0)  # should be [1, 0, 0]

    assert valueA[0] < 0.2
    assert valueA[1] > 0.8
    assert valueA[2] < 0.2

    assert valueB[0] < 0.2
    assert valueB[1] < 0.2
    assert valueB[2] > 0.8

    assert valueC[0] > 0.8
    assert valueC[1] < 0.2
    assert valueC[2] < 0.2


def test_routing_recurrency_compilation(Simulator, seed):
    D = 2
    model = spa.SPA(seed=seed)
    with model:
        model.buff1 = spa.Buffer(D, label="buff1")
        model.buff2 = spa.Buffer(D, label="buff2")
        actions = spa.Actions("0.5 --> buff2=buff1, buff1=buff2")
        model.bg = spa.BasalGanglia(actions)
        model.thal = spa.Thalamus(model.bg)

    with Simulator(model) as sim:
        assert sim


def test_nondefault_routing(Simulator, seed):
    D = 3
    model = spa.SPA(seed=seed)
    with model:
        model.ctrl = spa.Buffer(16, label="ctrl")

        def input_func(t):
            if t < 0.2:
                return "A"
            elif t < 0.4:
                return "B"
            else:
                return "C"

        model.input = spa.Input(ctrl=input_func)

        model.buff1 = spa.Buffer(D, label="buff1")
        model.buff2 = spa.Buffer(D, label="buff2")
        model.cmp = spa.Compare(D)

        node1 = nengo.Node([0, 1, 0])
        node2 = nengo.Node([0, 0, 1])

        nengo.Connection(node1, model.buff1.state.input)
        nengo.Connection(node2, model.buff2.state.input)

        actions = spa.Actions(
            "dot(ctrl, A) --> cmp_A=buff1, cmp_B=buff1",
            "dot(ctrl, B) --> cmp_A=buff1, cmp_B=buff2",
            "dot(ctrl, C) --> cmp_A=buff2, cmp_B=buff2",
        )
        model.bg = spa.BasalGanglia(actions)
        model.thal = spa.Thalamus(model.bg)

        compare_probe = nengo.Probe(model.cmp.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.6)

    similarity = sim.data[compare_probe]

    valueA = np.mean(similarity[150:200], axis=0)  # should be [1]
    valueB = np.mean(similarity[350:400], axis=0)  # should be [0]
    valueC = np.mean(similarity[550:600], axis=0)  # should be [1]

    assert valueA > 0.6
    assert valueB < 0.3
    assert valueC > 0.6


def test_errors():
    # motor does not exist
    with pytest.raises(SpaParseError):
        with spa.SPA() as model:
            model.vision = spa.Buffer(dimensions=16)
            actions = spa.Actions("0.5 --> motor=A")
            model.bg = spa.BasalGanglia(actions)

    # dot products not implemented
    with pytest.raises(NotImplementedError):
        with spa.SPA() as model:
            model.scalar = spa.Buffer(dimensions=16, subdimensions=1)
            actions = spa.Actions("0.5 --> scalar=dot(scalar, FOO)")
            model.bg = spa.BasalGanglia(actions)
            model.thalamus = spa.Thalamus(model.bg)



# Must be imported in order to register the build functions
from . import (
    connection,
    ensemble,
    learning_rules,
    network,
    neurons,
    node,
    probe,
    processes,
    transforms,
)
from .builder import Builder, Model
from .operator import Operator
from .signal import Signal



"""Expressions and Effects used to define all Actions."""

import re
import warnings

from nengo.exceptions import SpaParseError
from nengo.spa.action_objects import DotProduct, Source, Summation, Symbol


class Expression:
    """Parses an Action expression given a set of module outputs.

    Parameters
    ----------
    sources : list
        The names of the module outputs that can be used as part of the
        expression.
    expression : str
        The expression to evaluate. This either defines the utility of the
        action, or a value from an effect's assignment, given the state
        information from the module outputs. The simplest expression is
        ``"1"`` and they can get more complex, such as
        ``"0.5*(dot(vision, DOG) + dot(memory, CAT*MOUSE)*3 - 1)"``.
    """

    def __init__(self, sources, expression):
        self.objects = {}  # the list of known terms

        # make all the module outputs as known terms
        for name in sources:
            self.objects[name] = Source(name)
        # handle the term 'dot(a, b)' to mean DotProduct(a, b)
        self.objects["dot"] = DotProduct

        # use Python's eval to do the parsing of expressions for us
        self.validate_string(expression)
        sanitized_exp = " ".join(expression.split("\n"))
        try:
            self.expression = eval(sanitized_exp, {}, self)
        except NameError as e:
            raise SpaParseError(
                "Unknown module in expression '%s': %s" % (expression, e)
            )
        except TypeError as e:
            raise SpaParseError(
                "Invalid operator in expression '%s': %s" % (expression, e)
            )

        # normalize the result to a summation
        if not isinstance(self.expression, Summation):
            self.expression = Summation([self.expression])

    def validate_string(self, text):
        m = re.search("~[^a-zA-Z]", text)
        if m is not None:
            raise SpaParseError(
                "~ is only permitted before names (e.g., DOG) "
                "or modules (e.g., vision): %s" % text
            )

    def __getitem__(self, key):
        # this gets used by the eval in the constructor to create new
        # terms as needed
        item = self.objects.get(key, None)
        if item is None:
            if not key[0].isupper():
                raise SpaParseError(
                    "Semantic pointers must begin with a capital letter."
                )
            item = Symbol(key)
            self.objects[key] = item
        return item

    def __str__(self):
        return str(self.expression)


class Effect:
    """Parses an action effect given a set of module outputs.

    The following, in an `.Action` string, are valid effects::

        "motor=A"
        "motor=A*B, memory=vision+DOG"
        "motor=0.5*(memory*A + vision*B)"

    Parameters
    ----------
    sources : list
        The names of valid sources of information (SPA module outputs).
    sinks : list
        The names of valid places to send information (SPA module inputs).
    effect: str
        The action to implement. This is a set of assignment statements
        where the right hand side can be parsed with the `.Expression` class.
    """

    def __init__(self, sources, sinks, effect):
        self.effect = {}
        # Splits by ',' and separates into lvalue=rvalue. We cannot simply use
        # split, because the rvalue may contain commas in the case of dot(*,*).
        # However, *? is lazy, and * is greedy, making this regex work.
        for lvalue, rvalue in re.findall("(.*?)=([^=]*)(?:,|$)", effect):
            sink = lvalue.strip()
            if sink not in sinks:
                raise SpaParseError(
                    "Left-hand module '%s' from effect '%s=%s' "
                    "is not defined." % (lvalue, lvalue, rvalue)
                )
            if sink in self.effect:
                raise SpaParseError(
                    "Left-hand module '%s' from effect '%s=%s' "
                    "is assigned to multiple times in '%s'."
                    % (lvalue, lvalue, rvalue, effect)
                )
            self.effect[sink] = Expression(sources, rvalue)

    def __str__(self):
        return ", ".join("%s=%s" % x for x in self.effect.items())


class Action:
    """A single action.

    Consists of a conditional `.Expression` (optional) and an `.Effect`.

    Parameters
    ----------
    sources : list
        The names of valid sources of information (SPA module outputs).
    sinks : list
        The names of valid places to send information (SPA module inputs).
    action : str
        A string defining the action.  If ``'-->'`` is in the string, this
        is used as a marker to split the string into condition and effect.
        Otherwise it is treated as having no condition and just effect.
    name : str
        The name of this action.
    """

    def __init__(self, sources, sinks, action, name):
        self.name = name
        if "-->" in action:
            condition, effect = action.split("-->", 1)
            self.condition = Expression(sources, condition)
            self.effect = Effect(sources, sinks, effect)
        else:
            self.condition = None
            self.effect = Effect(sources, sinks, action)

    def __str__(self):
        return "<Action %s:\n  %s\n --> %s\n>" % (
            self.name,
            self.condition,
            self.effect,
        )


class Actions:
    """A collection of Action objects.

    The ``*args`` and ``**kwargs`` are treated as unnamed and named actions,
    respectively. The list of actions are only generated once
    `~.Actions.process` is called, since it needs access to the list of
    module inputs and outputs from the SPA object. The ``**kwargs`` are sorted
    alphabetically before being processed.
    """

    def __init__(self, *args, **kwargs):
        self.actions = None
        self.args = args
        self.kwargs = kwargs

    def add(self, *args, **kwargs):
        if self.actions is not None:
            warnings.warn(
                "The actions currently being added must be processed"
                " either by spa.BasalGanglia or spa.Cortical"
                " to be added to the model."
            )

        self.args += args
        self.kwargs.update(kwargs)

    @property
    def count(self):
        """Return the number of actions."""
        return len(self.args) + len(self.kwargs)

    def process(self, spa):
        """Parse the actions and generate the list of Action objects."""
        self.actions = []

        sources = list(spa.get_module_outputs())
        sinks = list(spa.get_module_inputs())

        sorted_kwargs = sorted(self.kwargs.items())

        for action in self.args:
            self.actions.append(Action(sources, sinks, action, name=None))
        for name, action in sorted_kwargs:
            self.actions.append(Action(sources, sinks, action, name=name))



import numpy as np

from . import numpy as npext  # pylint: disable = reimported


def tuning_curves(ens, sim, inputs=None):
    """Calculates the tuning curves of an ensemble.

    That is the neuron responses in dependence of the vector represented by the
    ensemble.

    For 1-dimensional ensembles, the unpacked return value of this function
    can be passed directly to ``matplotlib.pyplot.plot``.

    Parameters
    ----------
    ens : nengo.Ensemble
        Ensemble to calculate the tuning curves of.
    sim : nengo.Simulator
        Simulator providing information about the built ensemble. (An unbuilt
        ensemble does not have tuning curves assigned to it.)
    inputs : sequence of ndarray, optional
        The inputs at which the tuning curves will be evaluated. For each of
        the ``D`` ensemble dimensions one array of dimensionality ``D`` is needed.
        The output of :func:`numpy.meshgrid` with ``indexing='ij'`` is in the
        right format.

    Returns
    -------
    inputs : sequence of ndarray
        The passed or auto-generated ``inputs``.
    activities : ndarray
        The activities of the individual neurons given the ``inputs``.
        For ensembles with 1 dimension, the rows correspond to the ``inputs``
        and the columns to individual neurons.
        For ensembles with > 1 dimension, the last dimension enumerates the
        neurons, the remaining dimensions map to ``inputs``.

    See Also
    --------
    response_curves
    """
    # note: imported here to avoid circular imports
    from nengo.builder.ensemble import (  # pylint: disable=import-outside-toplevel
        get_activities,
    )

    if inputs is None:
        inputs = np.linspace(-ens.radius, ens.radius)
        if ens.dimensions > 1:
            inputs = npext.meshgrid_nd(*(ens.dimensions * [inputs]))
        else:
            inputs = [inputs]
        inputs = np.asarray(inputs).T
    else:
        inputs = np.asarray(inputs)

    eval_points = inputs.reshape((-1, ens.dimensions))
    activities = get_activities(sim.data[ens], ens, eval_points)
    return inputs, activities.reshape(inputs.shape[:-1] + (-1,))


def response_curves(ens, sim, inputs=None):
    """Calculates the response curves of an ensemble.

    That is the neuron responses in dependence of an already encoded value.
    This corresponds to the tuning curves along the neuron's preferred
    directions.

    Parameters
    ----------
    ens : nengo.Ensemble
        Ensemble to calculate the response curves of.
    sim : nengo.Simulator
        Simulator providing information about the built ensemble. (An unbuilt
        ensemble does not have response curves assigned to it.)
    inputs : 1d array, optional
        The inputs between -1 and 1 at which the neuron responses will be
        evaluated. They are assumed to be along each neuron's preferred
        direction.

    Returns
    -------
    inputs : 1d array
        The passed or auto-generated ``inputs``.
    activities : 2d array
        The activities of the individual neurons given the ``inputs``. The rows
        map to ``inputs`` and the columns to the neurons in the ensemble.

    See Also
    --------
    tuning_curves
    """

    if inputs is None:
        inputs = np.linspace(-1.0, 1.0)
    return inputs, ens.neuron_type.rates(inputs, sim.data[ens].gain, sim.data[ens].bias)


def _similarity(encoders, index, rows, cols=1):
    """Helper function to compute similarity for one encoder.

    Parameters
    ----------

    encoders: ndarray
        The encoders.
    index: int
        The encoder to compute for.
    rows: int
        The width of the 2d grid.
    cols: int
        The height of the 2d grid.
    """
    i = index % cols  # find the 2d location of the indexth element
    j = index // cols

    sim = 0  # total of dot products
    count = 0  # number of neighbours
    if i > 0:  # if we're not at the left edge, do the WEST comparison
        sim += np.dot(encoders[j * cols + i], encoders[j * cols + i - 1])
        count += 1
    if i < cols - 1:  # if we're not at the right edge, do EAST
        sim += np.dot(encoders[j * cols + i], encoders[j * cols + i + 1])
        count += 1
    if j > 0:  # if we're not at the top edge, do NORTH
        sim += np.dot(encoders[j * cols + i], encoders[(j - 1) * cols + i])
        count += 1
    if j < rows - 1:  # if we're not at the bottom edge, do SOUTH
        sim += np.dot(encoders[j * cols + i], encoders[(j + 1) * cols + i])
        count += 1
    return sim / count


def sorted_neurons(ensemble, sim, iterations=100, seed=None):
    """Sort neurons in an ensemble by encoder and intercept.

    Parameters
    ----------
    ensemble : Ensemble
        The population of neurons to be sorted.
    sim : Simulator
        Simulator providing information about the built ensemble.
    iterations: int
        The number of times to iterate during the sort.
    seed: float
        A random number seed.

    Returns
    -------
    indices: ndarray
        An array with sorted indices into the neurons in the ensemble

    Examples
    --------
    You can use this to generate an array of sorted indices for plotting. This
    can be done after collecting the data. E.g.

    .. testcode::

       from nengo.utils.ensemble import sorted_neurons
       from nengo.utils.matplotlib import rasterplot

       with nengo.Network() as net:
           ens = nengo.Ensemble(10, 1)
           ens_p = nengo.Probe(ens.neurons)

       with nengo.Simulator(net) as sim:
           sim.run(0.1)

           indices = sorted_neurons(ens, sim)
           rasterplot(sim.trange(), sim.data[ens_p][:, indices])

    .. testoutput::
       :hide:

       ...

    Notes
    -----
    The algorithm is for each encoder in the initial set, randomly
    pick another encoder and check to see if swapping those two
    encoders would reduce the average difference between the
    encoders and their neighbours.  Difference is measured as the
    dot product.  Each encoder has four neighbours (N, S, E, W),
    except for the ones on the edges which have fewer (no wrapping).
    This algorithm is repeated ``iterations`` times, so a total of
    ``iterations*N`` swaps are considered.
    """

    # Normalize all the encoders
    encoders = np.array(sim.data[ensemble].encoders)
    encoders /= npext.norm(encoders, axis=1, keepdims=True)

    # Make an array with the starting order of the neurons
    N = encoders.shape[0]
    indices = np.arange(N)
    rng = np.random.RandomState(seed)

    for _ in range(iterations):
        target = rng.randint(0, N, N)  # pick random swap targets
        for i in range(N):
            j = target[i]
            if i != j:  # if not swapping with yourself
                # compute similarity score how we are (unswapped)
                sim1 = _similarity(encoders, i, N) + _similarity(encoders, j, N)
                # swap the encoder
                encoders[[i, j], :] = encoders[[j, i], :]
                indices[[i, j]] = indices[[j, i]]
                # compute similarity score how we are (swapped)
                sim2 = _similarity(encoders, i, N) + _similarity(encoders, j, N)

                # if we were better unswapped
                if sim1 > sim2:
                    # swap them back
                    encoders[[i, j], :] = encoders[[j, i], :]
                    indices[[i, j]] = indices[[j, i]]

    return indices



import numpy as np
import pytest

import nengo
from nengo.exceptions import SimulationError, ValidationError
from nengo.rc import rc


def test_time(Simulator, allclose):
    with nengo.Network() as model:
        u = nengo.Node(output=lambda t: t)
        up = nengo.Probe(u)

    with Simulator(model) as sim:
        sim.run(1.0)

    t = sim.trange()
    x = sim.data[up].flatten()
    assert allclose(t, x, atol=1e-7, rtol=1e-4)


def test_simple(Simulator, plt, seed, allclose):
    m = nengo.Network(seed=seed)
    with m:
        input = nengo.Node(output=np.sin)
        p = nengo.Probe(input, "output")

    with Simulator(m) as sim:
        sim.run(0.5)

    plt.plot(sim.trange(), sim.data[p], label="sin")
    plt.legend(loc="best")

    sim_t = sim.trange()
    sim_in = sim.data[p].ravel()
    assert allclose(sim_in, np.sin(sim_t))


def test_connected(Simulator, plt, seed, allclose):
    m = nengo.Network(seed=seed)
    with m:
        input = nengo.Node(output=np.sin, label="input")
        output = nengo.Node(output=lambda t, x: np.square(x), size_in=1, label="output")
        nengo.Connection(input, output, synapse=None)  # Direct connection
        p_in = nengo.Probe(input, "output")
        p_out = nengo.Probe(output, "output")

    with Simulator(m) as sim:
        sim.run(0.5)

    t = sim.trange()
    plt.plot(t, sim.data[p_in], label="sin")
    plt.plot(t, sim.data[p_out], label="sin squared")
    plt.plot(t, np.sin(t), label="ideal sin")
    plt.plot(t, np.sin(t) ** 2, label="ideal squared")
    plt.legend(loc="best")

    sim_t = sim.trange()
    sim_sin = sim.data[p_in].ravel()
    sim_sq = sim.data[p_out].ravel()
    assert allclose(sim_sin, np.sin(sim_t))
    assert allclose(sim_sq, sim_sin ** 2)


def test_passthrough(Simulator, plt, seed, allclose):
    m = nengo.Network(seed=seed)
    with m:
        in1 = nengo.Node(output=np.sin)
        in2 = nengo.Node(output=lambda t: t)
        passthrough = nengo.Node(size_in=1)
        out = nengo.Node(output=lambda t, x: x, size_in=1)

        nengo.Connection(in1, passthrough, synapse=None)
        nengo.Connection(in2, passthrough, synapse=None)
        nengo.Connection(passthrough, out, synapse=None)

        in1_p = nengo.Probe(in1, "output")
        in2_p = nengo.Probe(in2, "output")
        out_p = nengo.Probe(out, "output")

    with Simulator(m) as sim:
        sim.run(0.5)

    plt.plot(sim.trange(), sim.data[in1_p] + sim.data[in2_p], label="in+in2")
    plt.plot(sim.trange()[:-2], sim.data[out_p][2:], label="out")
    plt.legend(loc="best")

    sim_in = sim.data[in1_p] + sim.data[in2_p]
    sim_out = sim.data[out_p]
    assert allclose(sim_in, sim_out)


def test_passthrough_filter(Simulator, plt, seed, allclose):
    m = nengo.Network(seed=seed)
    with m:
        omega = 2 * np.pi * 5
        u = nengo.Node(output=lambda t: np.sin(omega * t))
        passthrough = nengo.Node(size_in=1)
        v = nengo.Node(output=lambda t, x: x, size_in=1)

        synapse = nengo.Lowpass(0.3)
        nengo.Connection(u, passthrough, synapse=None)
        nengo.Connection(passthrough, v, synapse=synapse)

        up = nengo.Probe(u)
        vp = nengo.Probe(v)

    with Simulator(m) as sim:
        sim.run(1.0)

    t = sim.trange()
    x = sim.data[up]
    y = synapse.filt(x, dt=sim.dt, y0=0)
    z = sim.data[vp]

    plt.plot(t, x)
    plt.plot(t, y)
    plt.plot(t, z)

    assert allclose(y[:-1], z[1:], atol=1e-7, rtol=1e-4)


def test_circular(Simulator, seed, allclose):
    m = nengo.Network(seed=seed)
    with m:
        a = nengo.Node(output=lambda t, x: x + 1, size_in=1)
        b = nengo.Node(output=lambda t, x: x + 1, size_in=1)
        nengo.Connection(a, b, synapse=0)
        nengo.Connection(b, a, synapse=0)

        a_p = nengo.Probe(a, "output")
        b_p = nengo.Probe(b, "output")

    with Simulator(m) as sim:
        sim.run(0.5)

    assert allclose(sim.data[a_p], sim.data[b_p])


def test_outputparam_errors(Simulator):
    with nengo.Network() as model:
        # valid values
        nengo.Node(output=lambda t: t + 1)
        nengo.Node(output=0)
        nengo.Node(output=[0, 1])
        nengo.Node(output=nengo.processes.WhiteNoise())
        nengo.Node(size_in=1)

        # type errors
        with pytest.raises(ValidationError, match="Invalid node output type"):
            nengo.Node(output=object())
        with pytest.raises(ValidationError, match="is expected to accept exactly 1"):
            nengo.Node(output=np.add)

        # function errors
        nengo.Node(output=lambda t, x=[0]: t + 1, size_in=1)
        with pytest.raises(ValidationError, match="is expected to accept exactly 1"):
            nengo.Node(output=lambda t, x: x + 1)
        with pytest.raises(ValidationError, match="is expected to accept exactly 2"):
            nengo.Node(output=lambda t: t + 1, size_in=1)
        with pytest.raises(ValidationError, match="is expected to accept exactly 2"):
            nengo.Node(output=lambda t, x, y: t + 1, size_in=2)
        with pytest.raises(ValidationError, match="must be callable if size_in != 0"):
            nengo.Node(output=[0], size_in=1)
        with pytest.raises(ValidationError, match="must be callable if size_in != 0"):
            nengo.Node(output=0, size_in=1)

        # shape errors
        with pytest.raises(ValidationError, match=r"vector \(got shape \(2, 2\)\)"):
            nengo.Node(output=[[1, 2], [3, 4]])
        with pytest.raises(ValidationError, match=r"vector \(got shape \(1, 2\)\)"):
            nengo.Node(output=lambda t: [[t, t + 1]])
        with pytest.raises(ValidationError, match=r"vector \(got shape \(2, 2\)\)"):
            nengo.Node(output=[[3, 1], [2, 9]], size_out=4)
        with pytest.raises(ValidationError, match=r"\(5\) does not match size_out"):
            nengo.Node(output=[1, 2, 3, 4, 5], size_out=4)

    with Simulator(model):
        pass


def test_none(Simulator, seed):
    """Test for nodes that output None."""

    model = nengo.Network(seed=seed)

    # This function will fail, because at build time it will be
    # detected as producing output (func is called with 0 input)
    # but during the run it will produce None when t >=0.5
    def input_function(t):
        if t < 0.005:
            return [1]
        return None

    with model:
        u = nengo.Node(output=input_function)
        a = nengo.Ensemble(10, dimensions=1)
        nengo.Connection(u, a)

    with Simulator(model) as sim:
        with pytest.raises(SimulationError):
            sim.run(0.01)

    # This function will pass (with a warning), because it will
    # be determined at run time that the output function
    # returns None
    def none_function(t):
        pass

    model2 = nengo.Network()
    with model2:
        nengo.Node(output=none_function)

    with Simulator(model2) as sim:
        sim.run(0.01)


def test_unconnected_node(Simulator):
    """Make sure unconnected nodes still run."""
    hits = np.array(0)
    dt = 0.001

    def f(t):
        hits[...] += 1

    model = nengo.Network()
    with model:
        nengo.Node(f, size_in=0, size_out=0)
    with Simulator(model) as sim:
        assert hits == 0
        sim.run(dt)
        assert hits == 1
        sim.run(dt)
        assert hits == 2


def test_len():
    with nengo.Network():
        n1 = nengo.Node(None, size_in=1)
        n3 = nengo.Node([1, 2, 3])
        n4 = nengo.Node(lambda t: np.arange(4) * t)

    assert len(n1) == 1
    assert len(n3) == 3
    assert len(n4) == 4
    assert len(n1[0]) == 1
    assert len(n4[1:3]) == 2


def test_set_arraylike_output(Simulator):
    with nengo.Network() as model:
        # if output is None, size_out == size_in
        with pytest.warns(UserWarning):
            # warns since size_in != size_out and output is None
            passthrough = nengo.Node(None, size_in=20, size_out=30)
        assert passthrough.output is None
        assert passthrough.size_out == 20

        # if output is an array-like...
        # size_in must be 0
        with pytest.raises(ValidationError):
            nengo.Node(np.ones(1), size_in=1)
        # size_out must match
        with pytest.raises(ValidationError):
            nengo.Node(np.ones(3), size_out=2)
        # must be scalar or vector, not matrix
        with pytest.raises(ValidationError):
            nengo.Node(np.ones((2, 2)))
        # scalar gets promoted to float vector
        scalar = nengo.Node(2)
        assert scalar.output.shape == (1,)
        assert scalar.output.dtype == rc.float_dtype
        # vector stays 1D
        vector = nengo.Node(np.arange(3))
        assert vector.output.shape == (3,)
        assert vector.output.dtype == rc.float_dtype

    with Simulator(model):  # Ensure it all builds
        pass


def test_set_callable_output(Simulator):
    model = nengo.Network()

    counter = []

    def accumulate(t):
        counter.append(t)
        return t

    with model:
        # if we pass size_out, function should not be called
        assert len(counter) == 0
        accum_func = nengo.Node(accumulate, size_out=1)
        assert len(counter) == 0
        assert accum_func.size_out == 1

    def noreturn(t):
        pass

    with model:
        # if the function returns None, size_out == 0
        noreturn_func = nengo.Node(noreturn)
        assert noreturn_func.size_out == 0

    class TestObject:
        def __init__(self):
            self.val = 0

        def step(self, t):
            return self.val

    with model:
        # passing methods into nodes shouldn't cause validation errors
        test_obj = TestObject()
        nengo.Node(test_obj.step, size_out=1)

    with model:
        with pytest.raises(ValidationError):
            nengo.Node(lambda: 2.0, size_out=1)
        # if size_in is 0, should only take in t
        with pytest.raises(ValidationError):
            nengo.Node(lambda t, x: 2.0, size_in=0)
        with pytest.raises(ValidationError):
            nengo.Node(lambda t, x: 2.0, size_in=0, size_out=0)
        # if size_in > 0, should take both t and x
        with pytest.raises(ValidationError):
            nengo.Node(lambda t: t ** 2, size_in=1)
        # function must return a scalar or vector, not matrix
        with pytest.raises(ValidationError):
            nengo.Node(lambda t: np.ones((2, 2)))
        # variable length argument lists should be allowed (used in NengoSPA)
        nengo.Node(lambda *args: [2.0], size_in=0, size_out=1)
        nengo.Node(lambda *args: [2.0], size_in=1, size_out=1)

    with Simulator(model):  # Ensure it all builds
        pass


def test_delay(Simulator, plt):
    with nengo.Network() as model:
        a = nengo.Node(output=np.sin)
        b = nengo.Node(output=lambda t, x: -x, size_in=1)
        nengo.Connection(a[[0]], b, synapse=None)

        ap = nengo.Probe(a)
        bp = nengo.Probe(b)

    with Simulator(model) as sim:
        sim.run(0.005)

    plt.plot(sim.trange(), sim.data[ap])
    plt.plot(sim.trange(), -sim.data[bp])


def test_args(Simulator):
    class Fn:
        def __init__(self):
            self.last_x = None

        def __call__(self, t, x):
            assert isinstance(t, float)
            assert isinstance(x, np.ndarray)
            assert self.last_x is not x  # x should be a new copy on each call
            assert x.base is None  # x should not be a view (since it should be a copy)
            self.last_x = x
            assert x[0] == t

    with nengo.Network() as model:
        u = nengo.Node(lambda t: t)
        v = nengo.Node(Fn(), size_in=1, size_out=0)
        nengo.Connection(u, v, synapse=None)

    with Simulator(model) as sim:
        sim.run(0.01)


def test_wrong_output():
    """Setting a node as an input used to cause unbounded memory allocation."""

    with nengo.Network():
        node1 = nengo.Node(output=lambda t: t)

        with pytest.raises(ValueError):
            nengo.Node(node1)


def test_seed_error():
    """Setting a Node seed is currently not implemented."""
    with nengo.Network():
        with pytest.raises(NotImplementedError):
            nengo.Node(seed=1)


def test_node_with_offset_array_view(Simulator):
    v = np.array([[1.0, 2.0], [3.0, 4.0]])
    with nengo.Network() as model:
        node = nengo.Node(v[1])
        probe = nengo.Probe(node)
        assert probe

    with Simulator(model):
        pass


def test_node_with_unusual_strided_view(Simulator, seed):
    v = np.array([1.0, 2.0], dtype=complex)  # 16 byte itemsize
    with nengo.Network(seed=seed) as model:
        node = nengo.Node(v.real)  # 8 byte itemsize, but 16 byte strides
        probe = nengo.Probe(node)
        assert probe

    with Simulator(model):
        pass


@pytest.mark.parametrize("badval", [np.inf, np.nan, "string"])
def test_invalid_values(Simulator, badval):
    with nengo.Network() as model:
        with pytest.raises(ValidationError):
            node = nengo.Node(badval)

    with nengo.Network() as model:
        node = nengo.Node(lambda t: badval)
        ens = nengo.Ensemble(10, 1)
        nengo.Connection(node, ens)

    with Simulator(model) as sim:
        with pytest.raises(SimulationError):
            sim.run(0.01)


def test_probeable():
    with nengo.Network():
        assert nengo.Node(np.sin).probeable == ("output",)



import pytest

import nengo
from nengo.base import NengoObjectParam
from nengo.exceptions import ValidationError


def test_nengoobjectparam():
    """NengoObjectParam must be a Nengo object and is readonly by default."""

    class Test:
        nop = NengoObjectParam("nop")

    inst = Test()

    # Must be a Nengo object
    with pytest.raises(ValidationError):
        inst.nop = "a"

    # Can set it once
    a = nengo.Ensemble(10, dimensions=2, add_to_container=False)
    inst.nop = a.neurons
    assert inst.nop == a.neurons

    # Can't set it twice
    with pytest.raises(ValidationError):
        inst.nop = a


def test_nengoobjectparam_nonzero():
    """Can check that objects have nonzero size in/out."""

    class Test:
        n_in = NengoObjectParam("n_in", nonzero_size_in=True)
        n_out = NengoObjectParam("n_out", nonzero_size_out=True)

    inst = Test()
    with nengo.Network():
        n_in = nengo.Node(output=lambda t: t)
        n_out = nengo.Node(output=lambda t, x: None, size_in=1)
        probe = nengo.Probe(n_in)

        with pytest.raises(ValidationError):
            inst.n_in = n_in
        with pytest.raises(ValidationError):
            inst.n_out = n_out
        with pytest.raises(ValidationError):
            inst.n_out = probe

        inst.n_in = n_out
        inst.n_out = n_in



"""Functions for easy interactions with IPython and IPython notebooks."""

# pylint: disable=consider-using-f-string

import io

import numpy as np

try:
    import IPython
    from IPython import get_ipython
    from IPython.display import HTML

    if IPython.version_info[0] <= 3:  # pragma: no cover
        from IPython import nbformat
        from IPython.nbconvert import PythonExporter
    else:
        import nbformat
        from nbconvert import PythonExporter

except ImportError:  # pragma: no cover

    def get_ipython():
        return None


assert get_ipython


def check_ipy_version(min_version):
    """Check that ipython version is >= ``min_version``."""
    try:
        import IPython as ipy  # pylint: disable=import-outside-toplevel

        return ipy.version_info >= min_version
    except ImportError:  # pragma: no cover
        return False


def hide_input():
    """Hide the input of the Jupyter notebook input block this is executed in.

    Returns a link to toggle the visibility of the input block.
    """
    uuid = np.random.randint(np.iinfo(np.int32).max)

    script = """
        <a id="%(uuid)s" href="javascript:toggle_input_%(uuid)s()"
          >Show Input</a>

        <script type="text/javascript">
        var toggle_input_%(uuid)s;
        (function() {
            if (typeof jQuery == 'undefined') {
                // no jQuery
                var link_%(uuid)s = document.getElementById("%(uuid)s");
                var cell = link_%(uuid)s;
                while (cell.className.split(' ')[0] != "cell"
                       && cell.className.split(' ')[0] != "nboutput") {
                    cell = cell.parentNode;
                }
                var input_%(uuid)s;
                if (cell.className.split(' ')[0] == "cell") {
                    for (var i = 0; i < cell.children.length; i++) {
                        if (cell.children[i].className.split(' ')[0]
                            == "input") {
                            input_%(uuid)s = cell.children[i];
                        }
                    }
                } else {
                    input_%(uuid)s = cell.previousElementSibling;
                }
                input_%(uuid)s.style.display = "none"; // hide

                toggle_input_%(uuid)s = function() {
                    if (input_%(uuid)s.style.display == "none") {
                        input_%(uuid)s.style.display = ""; // show
                        link_%(uuid)s.innerHTML = "Hide Input";
                    } else {
                        input_%(uuid)s.style.display = "none"; // hide
                        link_%(uuid)s.innerHTML = "Show Input";
                    }
                }

            } else {
                // jQuery
                var link_%(uuid)s = $("a[id='%(uuid)s']");
                var cell_%(uuid)s = link_%(uuid)s.parents("div.cell:first");
                if (cell_%(uuid)s.length == 0) {
                    cell_%(uuid)s = link_%(uuid)s.parents(
                        "div.nboutput:first");
                }
                var input_%(uuid)s = cell_%(uuid)s.children("div.input");
                if (input_%(uuid)s.length == 0) {
                    input_%(uuid)s = cell_%(uuid)s.prev("div.nbinput");
                }
                input_%(uuid)s.hide();

                toggle_input_%(uuid)s = function() {
                    if (input_%(uuid)s.is(':hidden')) {
                        input_%(uuid)s.slideDown();
                        link_%(uuid)s[0].innerHTML = "Hide Input";
                    } else {
                        input_%(uuid)s.slideUp();
                        link_%(uuid)s[0].innerHTML = "Show Input";
                    }
                }
            }
        }());
        </script>
    """ % dict(
        uuid=uuid
    )

    return HTML(script)


def load_notebook(nb_path):
    """Load notebook from file."""
    with io.open(nb_path, "r", encoding="utf-8") as f:
        nb = nbformat.read(f, as_version=4)
    return nb


def export_py(nb, dest_path=None):
    """Convert notebook to Python script.

    Optionally saves script to dest_path.
    """
    exporter = PythonExporter()
    body, _ = exporter.from_notebook_node(nb)

    # Remove all lines with get_ipython
    while "get_ipython()" in body:
        ind0 = body.find("get_ipython()")
        ind1 = body.find("\n", ind0)
        body = body[:ind0] + body[(ind1 + 1) :]

    if dest_path is not None:
        with io.open(dest_path, "w", encoding="utf-8") as f:
            f.write(body)
    return body


def iter_cells(nb, cell_type="code"):
    """Iterate over cells of a notebok."""
    return (cell for cell in nb.cells if cell.cell_type == cell_type)



import numpy as np

from nengo.builder.builder import Builder
from nengo.builder.operator import Operator
from nengo.builder.signal import Signal
from nengo.exceptions import BuildError
from nengo.neurons import NeuronType, RatesToSpikesNeuronType
from nengo.utils.numpy import is_array_like


class SimNeurons(Operator):
    """Set a neuron model output for the given input current.

    Implements ``neurons.step(dt, J, **state)``.

    Parameters
    ----------
    neurons : NeuronType
        The `.NeuronType`, which defines a ``step`` function.
    J : Signal
        The input current.
    output : Signal
        The neuron output signal that will be set
    state : list, optional
        A list of additional neuron state signals set by ``step``.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    J : Signal
        The input current.
    neurons : NeuronType
        The `.NeuronType`, which defines a ``step`` function.
    output : Signal
        The neuron output signal that will be set.
    state : list
        A list of additional neuron state signals set by ``step``.
    tag : str or None
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[output] + state``
    2. incs ``[]``
    3. reads ``[J]``
    4. updates ``[]``
    """

    def __init__(self, neurons, J, output, state=None, tag=None):
        super().__init__(tag=tag)
        self.neurons = neurons

        self.sets = [output]
        self.incs = []
        self.reads = [J]
        self.updates = []

        self.state_idxs = {}
        self.state_extra = {}
        if state is not None:
            for name, obj in state.items():
                if isinstance(obj, Signal):
                    # The signals actually stored in `self.sets` can be modified by the
                    # optimizer. To allow this possibility, we store the index of the
                    # signal in the sets list instead of storing the signal itself.
                    self.state_idxs[name] = len(self.sets)
                    self.sets.append(obj)
                else:
                    # The only supported extra is a RandomState right now
                    assert isinstance(obj, np.random.RandomState)
                    self.state_extra[name] = obj

    @property
    def J(self):
        return self.reads[0]

    @property
    def output(self):
        return self.sets[0]

    @property
    def state(self):
        return {key: self.sets[idx] for key, idx in self.state_idxs.items()}

    @property
    def _descstr(self):
        return f"{self.neurons}, {self.J}, {self.output}"

    def make_step(self, signals, dt, rng):
        J = signals[self.J]
        output = signals[self.output]
        state = {name: signals[sig] for name, sig in self.state.items()}
        state.update(self.state_extra)

        def step_simneurons():
            self.neurons.step(dt, J, output, **state)

        return step_simneurons


@Builder.register(NeuronType)
def build_neurons(model, neurontype, neurons, input_sig=None, output_sig=None):
    """Builds a `.NeuronType` object into a model.

    This function adds a `.SimNeurons` operator connecting the input current to the
    neural output signals, and handles any additional state variables defined
    within the neuron type.

    Parameters
    ----------
    model : Model
        The model to build into.
    neurontype : NeuronType
        Neuron type to build.
    neuron : Neurons
        The neuron population object corresponding to the neuron type.

    Notes
    -----
    Does not modify ``model.params[]`` and can therefore be called
    more than once with the same `.NeuronType` instance.
    """
    input_sig = model.sig[neurons]["in"] if input_sig is None else input_sig
    output_sig = model.sig[neurons]["out"] if output_sig is None else output_sig

    dtype = input_sig.dtype
    n_neurons = neurons.size_in
    rng = np.random.RandomState(model.seeds[neurons.ensemble] + 1)
    state_init = neurontype.make_state(n_neurons, rng=rng, dtype=dtype)
    state = {}

    for key, init in state_init.items():
        if key in model.sig[neurons]:
            raise BuildError(f"State name '{key}' overlaps with existing signal name")
        if is_array_like(init):
            model.sig[neurons][key] = Signal(
                initial_value=init, name=f"{neurons}.{key}"
            )
            state[key] = model.sig[neurons][key]
        elif isinstance(init, np.random.RandomState):
            # Pass through RandomState instances
            state[key] = init
        else:
            raise BuildError(
                f"State '{key}' is of type '{type(init).__name__}'. Only array-likes "
                "and RandomStates are currently supported."
            )

    model.add_op(
        SimNeurons(neurons=neurontype, J=input_sig, output=output_sig, state=state)
    )


@Builder.register(RatesToSpikesNeuronType)
def build_rates_to_spikes(model, neurontype, neurons):
    """Builds a `.RatesToSpikesNeuronType` object into a model.

    This function adds two `.SimNeurons` operators. The first one handles
    simulating the base_type, converting input signals into rates. The second
    one takes those rates as input and emits spikes.

    Parameters
    ----------
    model : Model
        The model to build into.
    neurontype : RatesToSpikesNeuronType
        Neuron type to build.
    neuron : Neurons
        The neuron population object corresponding to the neuron type.

    Notes
    -----
    Does not modify ``model.params[]`` and can therefore be called
    more than once with the same `.NeuronType` instance.
    """

    in_sig = model.sig[neurons]["in"]
    out_sig = model.sig[neurons]["out"]
    rate_sig = Signal(shape=model.sig[neurons]["in"].shape, name=f"{neurons}.rate_out")
    model.sig[neurons]["rate_out"] = rate_sig

    # build the base neuron type
    build_neurons(
        model, neurontype.base_type, neurons, input_sig=in_sig, output_sig=rate_sig
    )

    # apply the RatesToSpikes conversion to output of base neuron type
    build_neurons(model, neurontype, neurons, input_sig=rate_sig, output_sig=out_sig)



from io import BytesIO

import pytest

import nengo
import nengo.spa
from nengo.exceptions import (
    BuildError,
    CacheIOError,
    ConfigError,
    ConvergenceError,
    FingerprintError,
    MovedError,
    NetworkContextError,
    ObsoleteError,
    ReadonlyError,
    SignalError,
    SimulationError,
    SimulatorClosed,
    SpaModuleError,
    SpaParseError,
    Unconvertible,
    ValidationError,
)
from nengo.params import ObsoleteParam
from nengo.rc import RC_DEFAULTS, rc
from nengo.utils.builder import generate_graphviz


def check_tb_entries(entries, name_statement_pairs):
    assert len(entries) == len(name_statement_pairs)
    for entry, (name, statement) in zip(entries, name_statement_pairs):
        assert entry.name == name
        if statement.endswith("..."):
            assert str(entry.statement.deindent()).startswith(statement[:-3])
        else:
            assert str(entry.statement.deindent()) == statement


def test_validation_error(request):
    # Ensure settings are set back to default after the test, even if it fails
    request.addfinalizer(
        lambda: rc.set(
            "exceptions", "simplified", str(RC_DEFAULTS["exceptions"]["simplified"])
        )
    )

    nengo.rc["exceptions"]["simplified"] = "False"

    with nengo.Network():
        with pytest.raises(ValidationError) as excinfo:
            nengo.Ensemble(n_neurons=0, dimensions=1)

    assert str(excinfo.value) == (
        "Ensemble.n_neurons: Value must be greater than or equal to 1 (got 0)"
    )
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_validation_error", "nengo.Ensemble(n_neurons=0, dimensions=1)"),
            ("__call__", "inst.__init__(*args, **kwargs)"),
            ("__init__", "self.n_neurons = n_neurons"),
            ("__setattr__", "super().__setattr__(name, val)"),
            ("__setattr__", "super().__setattr__(name, val)"),
            ("__set__", "self.data[instance] = self.coerce(instance, value)"),
            ("coerce", "return super().coerce(instance, num)"),
            ("coerce", "raise ValidationError(..."),
        ],
    )

    nengo.rc["exceptions"]["simplified"] = "True"

    with pytest.raises(ValidationError) as excinfo:
        nengo.dists.PDF(x=[1, 1], p=[0.1, 0.2])

    assert str(excinfo.value) == "PDF.p: PDF must sum to one (sums to 0.300000)"
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_validation_error", "nengo.dists.PDF(x=[1, 1], p=[0.1, 0.2])"),
            ("__init__", "raise ValidationError(..."),
        ],
    )


def test_convergence_error():
    with pytest.raises(ConvergenceError) as excinfo:
        nengo.dists.QuasirandomSequence()._phi(1, tol=0)
    assert str(excinfo.value) == "'phi' computation did not converge for d=1"
    check_tb_entries(
        excinfo.traceback,
        [
            (
                "test_convergence_error",
                "nengo.dists.QuasirandomSequence()._phi(1, tol=0)",
            ),
            ("_phi", "raise ConvergenceError(..."),
        ],
    )


def test_readonly_error():
    with nengo.Network():
        ens = nengo.Ensemble(n_neurons=10, dimensions=1)
        p = nengo.Probe(ens)
        with pytest.raises(ReadonlyError) as excinfo:
            p.target = ens
    assert (
        str(excinfo.value) == "Probe.target: target is read-only and cannot be changed"
    )

    class Frozen(nengo.params.FrozenObject):
        p = nengo.params.Parameter("p", readonly=False)

    with pytest.raises(ReadonlyError) as excinfo:
        Frozen()

    assert str(excinfo.value).endswith(
        "All parameters of a FrozenObject must be readonly"
    )
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_readonly_error", "Frozen()"),
            (
                "__init__",
                "raise ReadonlyError(attr=p, obj=self, msg=msg)",
            ),
        ],
    )


def test_build_error():
    model = nengo.builder.Model()
    with pytest.raises(BuildError) as excinfo:
        nengo.builder.Builder.build(model, "")

    assert str(excinfo.value) == "Cannot build object of type 'str'"
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_build_error", 'nengo.builder.Builder.build(model, "")'),
            (
                "build",
                'raise BuildError(f"Cannot build object of '
                "type '{type(obj).__name__}'\")",
            ),
        ],
    )


def test_obsolete_error():
    class Test:
        ab = ObsoleteParam("ab", "msg")

    inst = Test()
    with pytest.raises(ObsoleteError) as excinfo:
        print(inst.ab)

    assert str(excinfo.value) == "Obsolete: msg"
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_obsolete_error", "print(inst.ab)"),
            ("__get__", "self.raise_error()"),
            (
                "raise_error",
                "raise ObsoleteError(self.short_msg, since=self.since, url=self.url)",
            ),
        ],
    )


def test_moved_error():
    with pytest.raises(MovedError) as excinfo:
        generate_graphviz()

    assert str(excinfo.value) == "This feature has been moved to nengo_extras.graphviz"
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_moved_error", "generate_graphviz()"),
            ("generate_graphviz", 'raise MovedError(location="nengo_extras.graphviz")'),
        ],
    )


def test_config_error():
    with pytest.raises(ConfigError) as excinfo:
        print(nengo.Network().config[object])

    assert str(excinfo.value) == (
        "Type 'object' is not set up for configuration. "
        "Call 'configures(object)' first."
    )
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_config_error", "print(nengo.Network().config[object])"),
            ("__getitem__", "raise ConfigError(..."),
        ],
    )


def test_spa_module_error():
    with pytest.raises(SpaModuleError) as excinfo:
        with nengo.spa.SPA():
            nengo.spa.State(1, label="1")

    assert (
        str(excinfo.value) == '<State "1"> must be set as an attribute of a SPA network'
    )
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_spa_module_error", 'nengo.spa.State(1, label="1")'),
            ("__exit__", "raise SpaModuleError(..."),
        ],
    )


def test_spa_parse_error():
    vocab = nengo.spa.Vocabulary(16)
    with pytest.raises(SpaParseError) as excinfo:
        print(vocab["a"])

    assert str(excinfo.value) == "Semantic pointers must begin with a capital letter."
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_spa_parse_error", 'print(vocab["a"])'),
            (
                "__getitem__",
                'raise SpaParseError("Semantic pointers must begin with a capital '
                'letter.")',
            ),
        ],
    )


def test_simulator_closed():
    with nengo.Network() as net:
        nengo.Ensemble(10, 1)
    with nengo.Simulator(net) as sim:
        sim.run(0.01)
    with pytest.raises(SimulatorClosed) as excinfo:
        sim.run(0.01)

    assert str(excinfo.value) == "Simulator cannot run because it is closed."
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_simulator_closed", "sim.run(0.01)"),
            ("run", "self.run_steps(steps, progress_bar=progress_bar)"),
            ("run_steps", "self.step()"),
            (
                "step",
                'raise SimulatorClosed("Simulator cannot run because it is closed.")',
            ),
        ],
    )


def test_simulation_error():
    with nengo.Network() as net:
        nengo.Node(lambda t: None if t > 0.002 else 1.0)
    with nengo.Simulator(net) as sim:
        with pytest.raises(SimulationError) as excinfo:
            sim.run(0.003)

    assert str(excinfo.value) == "Function '<lambda>' returned non-finite value"
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_simulation_error", "sim.run(0.003)"),
            ("run", "self.run_steps(steps, progress_bar=progress_bar)"),
            ("run_steps", "self.step()"),
            ("step", "step_fn()"),
            ("step_simpyfunc", "raise SimulationError(..."),
        ],
    )


def test_signal_error():
    s = nengo.builder.signal.Signal([1])
    with pytest.raises(SignalError) as excinfo:
        s.initial_value = 0

    assert str(excinfo.value) == "Cannot change initial value after initialization"
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_signal_error", "s.initial_value = 0"),
            (
                "initial_value",
                'raise SignalError("Cannot change initial value after initialization")',
            ),
        ],
    )


def test_fingerprint_error():
    with pytest.raises(FingerprintError) as excinfo:
        nengo.cache.Fingerprint(lambda x: x)

    assert str(excinfo.value) == "Object of type 'function' cannot be fingerprinted."
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_fingerprint_error", "nengo.cache.Fingerprint(lambda x: x)"),
            ("__init__", "raise FingerprintError(..."),
        ],
    )


def test_network_context_error(request):
    request.addfinalizer(nengo.Network.context.clear)

    with pytest.raises(NetworkContextError) as excinfo:
        with nengo.Network(label="net"):
            nengo.Network.context.append("bad")
            nengo.Ensemble(10, 1)

    assert str(excinfo.value) == (
        "Network.context in bad state; was expecting current context to be "
        "'<Network \"net\">' but instead got 'bad'."
    )
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_network_context_error", "nengo.Ensemble(10, 1)"),
            ("__exit__", "raise NetworkContextError(..."),
        ],
    )


def test_unconvertible():
    with nengo.Network() as net:
        n = nengo.Node(output=None, size_in=1)
        nengo.Connection(n, n, synapse=None)

    with pytest.raises(Unconvertible) as excinfo:
        nengo.utils.builder.remove_passthrough_nodes(net.nodes, net.connections)

    assert str(excinfo.value) == "Cannot remove a Node with a feedback connection"
    check_tb_entries(
        excinfo.traceback,
        [
            (
                "test_unconvertible",
                "nengo.utils.builder.remove_passthrough_nodes(net.nodes, "
                "net.connections)",
            ),
            ("remove_passthrough_nodes", "raise Unconvertible(..."),
        ],
    )


def test_cache_io_error():
    bio = BytesIO(b"a" * 40)
    with pytest.raises(CacheIOError) as excinfo:
        nengo.utils.nco.read(bio)

    assert str(excinfo.value) == "Not a Nengo cache object file."
    check_tb_entries(
        excinfo.traceback,
        [
            ("test_cache_io_error", "nengo.utils.nco.read(bio)"),
            ("read", 'raise CacheIOError("Not a Nengo cache object file.")'),
        ],
    )



from nengo.builder.builder import Builder
from nengo.builder.connection import slice_signal
from nengo.builder.operator import Copy, Operator, Reset
from nengo.builder.signal import Signal
from nengo.connection import Connection, LearningRule
from nengo.ensemble import Ensemble, Neurons
from nengo.exceptions import BuildError
from nengo.node import Node
from nengo.probe import Probe


class SimProbe(Operator):
    """Mark a signal as being probed.

    This performs no computations, but marks ``signal`` as being read. This is
    necessary for the rare case in which a node with constant output is probed
    directly without that signal being otherwise used.

    Parameters
    ----------
    signal : Signal
        The probed signal.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[]``
    3. reads ``[signal]``
    4. updates ``[]``
    """

    def __init__(self, signal, tag=None):
        super().__init__(tag=tag)
        self.sets = []
        self.incs = []
        self.reads = [signal]
        self.updates = []

    @property
    def signal(self):
        """The probed signal"""
        return self.reads[0]

    def make_step(self, signals, dt, rng):
        def step():
            pass

        return step


def conn_probe(model, probe):
    """Build a "connection" probe type.

    Connection probes create a connection from the target, and probe
    the resulting signal (used when you want to probe the default
    output of an object, which may not have a predefined signal).
    """

    conn = Connection(
        probe.target,
        probe,
        synapse=probe.synapse,
        solver=probe.solver,
        add_to_container=False,
    )

    # Set connection's seed to probe's (which isn't used elsewhere)
    model.seeded[conn] = model.seeded[probe]
    model.seeds[conn] = model.seeds[probe]

    # Make a sink signal for the connection
    model.sig[probe]["in"] = Signal(shape=conn.size_out, name=str(probe))
    model.add_op(Reset(model.sig[probe]["in"]))

    # Build the connection
    model.build(conn)


def signal_probe(model, key, probe):
    """Build a "signal" probe type.

    Signal probes directly probe a target signal.
    """

    try:
        sig = model.sig[probe.obj][key]
    except (IndexError, KeyError) as e:
        raise BuildError(f"Attribute '{key}' is not probeable on {probe.obj}.") from e

    if sig is None:
        raise BuildError(f"Attribute '{key}' on {probe.obj} is None, cannot be probed")

    if probe.slice is not None:
        sig = slice_signal(model, sig, probe.slice)

    if probe.synapse is None:
        model.sig[probe]["in"] = sig
    else:
        model.sig[probe]["in"] = Signal(shape=sig.shape, name=str(probe))
        model.sig[probe]["filtered"] = model.build(probe.synapse, sig, mode="update")
        model.add_op(Copy(model.sig[probe]["filtered"], model.sig[probe]["in"]))


probemap = {
    Ensemble: {"decoded_output": None, "input": "in", "scaled_encoders": "encoders"},
    Neurons: {"output": "out", "spikes": "out", "rates": "out", "input": "in"},
    Node: {"output": "out"},
    Connection: {"output": "weighted", "input": "in"},
    LearningRule: {},  # make LR signals probeable, but no mapping required
}


@Builder.register(Probe)
def build_probe(model, probe):
    """Builds a `.Probe` object into a model.

    Under the hood, there are two types of probes:
    connection probes and signal probes.

    Connection probes are those that are built by creating a new `.Connection`
    object from the probe's target to the probe, and calling that connection's
    build function. Creating and building a connection ensure that the result
    of probing the target's attribute is the same as would result from that
    target being connected to another object.

    Signal probes are those that are built by finding the correct `.Signal`
    in the model and calling the build function corresponding to the probe's
    synapse.

    Parameters
    ----------
    model : Model
        The model to build into.
    probe : Probe
        The connection to build.

    Notes
    -----
    Sets ``model.params[probe]`` to a list.
    `.Simulator` appends to that list when running a simulation.
    """

    # find the right parent class in `objtypes`, using `isinstance`
    for nengotype, probeables in probemap.items():
        if isinstance(probe.obj, nengotype):
            key = probeables.get(probe.attr, probe.attr)
            break
    else:
        raise BuildError(f"Type '{type(probe.obj).__name__}' is not probeable")

    if key is None:
        conn_probe(model, probe)
    else:
        signal_probe(model, key, probe)

    model.add_op(SimProbe(model.sig[probe]["in"]))
    model.probes.append(probe)

    # Simulator will fill this list with probe data during simulation
    model.params[probe] = []



import sys
import time
from collections import namedtuple

import pytest

from nengo import rc
from nengo.exceptions import ValidationError
from nengo.utils.progress import (
    AutoProgressBar,
    HtmlProgressBar,
    NoProgressBar,
    Progress,
    ProgressBar,
    ProgressTracker,
    TerminalProgressBar,
    WriteProgressToFile,
    get_default_progressbar,
    to_progressbar,
)

Update = namedtuple(
    "Update", ("name_during", "name_after", "n_steps", "max_steps", "finished")
)


class ProgressBarMock(ProgressBar):
    def __init__(self):
        super().__init__()
        self.n_update_calls = 0
        self.updates = []
        self.closed = False

    def update(self, progress):
        self.n_update_calls += 1
        self.updates.append(
            Update(
                name_during=progress.name_during,
                name_after=progress.name_after,
                n_steps=progress.n_steps,
                max_steps=progress.max_steps,
                finished=progress.finished,
            )
        )

    def close(self):
        self.closed = True


class TestProgress:
    def test_progress_calculation(self):
        with Progress(max_steps=10) as p:
            assert p.progress == 0.0
            for _ in range(5):
                p.step()
            assert p.progress == 0.5
            p.step(5)
            assert p.progress == 1.0

    def test_finished_property(self):
        with Progress(max_steps=10) as p:
            assert not p.finished
            p.step(5)
            assert not p.finished
        assert p.finished

    def test_success_property(self):
        with Progress(max_steps=10) as p:
            assert p.success is None
        assert p.success

        try:
            with Progress(max_steps=10) as p2:
                raise UnboundLocalError()
        except UnboundLocalError:
            pass
        assert not p2.success

    def test_elapsed_seconds(self, monkeypatch):
        t = 1.0
        monkeypatch.setattr(time, "time", lambda: t)

        with Progress(max_steps=10) as p:
            t = 10.0

        assert p.elapsed_seconds() == 9.0

    def test_eta(self):
        with Progress(max_steps=10) as p:
            assert p.eta() == -1  # no estimate available yet
            p.step()
            assert p.eta() >= 0.0

    def test_max_steps(self):
        with pytest.raises(ValidationError):
            with Progress(max_steps=0):
                pass

        with pytest.raises(ValidationError):
            with Progress(max_steps=-1):
                pass

    def test_unknown_number_of_steps(self, monkeypatch):
        t = 1.0
        monkeypatch.setattr(time, "time", lambda: t)

        with Progress() as p:
            p.step()
            t = 10.0
            assert p.progress == 0.0
            assert p.eta() == -1

        assert p.n_steps == 1
        assert p.elapsed_seconds() == 9.0


class TestAutoProgressBar:
    class ProgressMock(Progress):
        def __init__(self, eta):
            super().__init__()
            self.eta = lambda: eta
            # make this 0 to easily control `long_eta` in AutoProgressBar.update
            self.elapsed_seconds = lambda: 0
            # make this small so `min_delay < time.time()` in AutoProgressBar.update
            self.time_start = 1.0

    def test_progress_not_shown_if_eta_below_threshold(self):
        progress_mock = self.ProgressMock(0.2)
        progress_bar = ProgressBarMock()
        auto_progress = AutoProgressBar(progress_bar, min_eta=10.0)

        for _ in range(10):
            auto_progress.update(progress_mock)

        assert progress_bar.n_update_calls == 0

    def test_progress_shown_if_eta_above_threshold(self):
        progress_mock = self.ProgressMock(20)
        progress_bar = ProgressBarMock()
        auto_progress = AutoProgressBar(progress_bar, min_eta=10.0)

        for _ in range(10):
            auto_progress.update(progress_mock)

        assert progress_bar.n_update_calls >= 10

    def test_progress_shown_when_finished(self):
        progress_mock = self.ProgressMock(0.2)
        progress_bar = ProgressBarMock()
        auto_progress = AutoProgressBar(progress_bar, min_eta=10.0)

        auto_progress.update(progress_mock)
        assert progress_bar.n_update_calls == 0
        progress_mock.finished = True
        auto_progress.update(progress_mock)
        assert progress_bar.n_update_calls >= 1

    def test_close(self):
        progress_bar = ProgressBarMock()
        auto_progress = AutoProgressBar(progress_bar)
        auto_progress.close()
        assert progress_bar.closed

    def test_get_default_progressbar(self):
        rc["progress"]["progress_bar"] = "False"
        assert isinstance(get_default_progressbar(), NoProgressBar)

        rc["progress"]["progress_bar"] = "True"
        assert isinstance(get_default_progressbar(), AutoProgressBar)

        rc["progress"]["progress_bar"] = "nengo.utils.progress.HtmlProgressBar"
        assert isinstance(get_default_progressbar(), HtmlProgressBar)

        rc["progress"]["progress_bar"] = "nengo.utils.progress.TerminalProgressBar"
        assert isinstance(get_default_progressbar(), TerminalProgressBar)


def test_write_progress_to_file(tmp_path):
    """Tests the WriteProgressToFile progress bar type"""

    def check_file(filename, startstring):
        with open(filename, "r", encoding="utf-8") as fh:
            data = fh.read()
        assert data.startswith(startstring)

    filename = tmp_path / "test_write_progress_file.txt"
    progress = Progress(name_during="myprog", max_steps=2)
    bar = WriteProgressToFile(filename)

    with progress:
        bar.update(progress)
        check_file(filename, "0%, ETA")

        progress.step()
        bar.update(progress)
        check_file(filename, "50%, ETA")

        progress.step()
        bar.update(progress)
        check_file(filename, "100%, ETA")

    bar.update(progress)
    check_file(filename, "myprog finished in")


@pytest.mark.skipif(
    sys.platform == "win32", reason="Threading in Windows is less reliable"
)
def test_progress_tracker():
    update_interval = 0.001
    sleep_interval = 20 * update_interval
    stages = 4
    steps = 3
    total_progress = Progress(name_during="total_prog", max_steps=stages)
    progress_bar = ProgressBarMock()
    tracker = ProgressTracker(
        progress_bar, total_progress, update_interval=update_interval
    )

    assert not progress_bar.closed
    with tracker:
        for i in range(stages):
            sub_progress = tracker.next_stage(name_during=f"stage{i}", max_steps=steps)
            with sub_progress:
                for j in range(steps):
                    sub_progress.step()
                    time.sleep(sleep_interval)
                    assert progress_bar.updates[-1].name_during == f"stage{i}"
                    assert progress_bar.updates[-1].n_steps == j + 1

            time.sleep(sleep_interval)
            assert progress_bar.updates[-1].name_during == "total_prog"
            assert progress_bar.updates[-1].n_steps == i

    assert progress_bar.closed


def test_to_progress_bar(request, tmp_path):
    def finalizer(val=rc["progress"]["progress_bar"]):
        rc["progress"]["progress_bar"] = val

    request.addfinalizer(finalizer)

    # test false values
    for val in (None, False):
        assert isinstance(to_progressbar(val), NoProgressBar)

    # test true value
    rc["progress"]["progress_bar"] = "nengo.utils.progress.HtmlProgressBar"
    assert isinstance(to_progressbar(True), HtmlProgressBar)

    progress_bar = WriteProgressToFile(tmp_path / "dummyfile.txt")
    assert to_progressbar(progress_bar) is progress_bar



import logging

import numpy as np
import pytest

import nengo
from nengo.exceptions import MovedError, Unconvertible, ValidationError
from nengo.transforms import NoTransform
from nengo.utils.builder import (
    _create_replacement_connection,
    full_transform,
    generate_graphviz,
    objs_and_connections,
    remove_passthrough_nodes,
)


def test_full_transform():
    """Tests ``full_transform`` and its exceptions"""
    N = 30

    with nengo.Network():
        neurons3 = nengo.Ensemble(3, dimensions=1).neurons
        ens1 = nengo.Ensemble(N, dimensions=1)
        ens2 = nengo.Ensemble(N, dimensions=2)
        ens3 = nengo.Ensemble(N, dimensions=3)
        node1 = nengo.Node(output=[0])
        node2 = nengo.Node(output=[0, 0])
        node3 = nengo.Node(output=[0, 0, 0])

        # error for non-Dense transform
        conn = nengo.Connection(
            ens2, ens3, transform=nengo.transforms.Sparse((3, 2), indices=[(0, 0)])
        )
        with pytest.raises(ValidationError, match="can only be applied to Dense"):
            full_transform(conn)

        # Pre slice with default transform -> 1x3 transform
        conn = nengo.Connection(node3[2], ens1)
        assert isinstance(conn.transform, NoTransform)
        assert np.all(full_transform(conn) == np.array([[0, 0, 1]]))

        # Post slice with 1x1 transform -> 1x2 transform
        conn = nengo.Connection(node2[0], ens1, transform=-2)
        assert np.all(conn.transform.init == np.array(-2))
        assert np.all(full_transform(conn) == np.array([[-2, 0]]))

        # Post slice with 2x1 tranfsorm -> 3x1 transform
        conn = nengo.Connection(node1, ens3[::2], transform=[[1], [2]])
        assert np.all(conn.transform.init == np.array([[1], [2]]))
        assert np.all(full_transform(conn) == np.array([[1], [0], [2]]))

        # Both slices with 2x1 transform -> 3x2 transform
        conn = nengo.Connection(ens2[-1], neurons3[1:], transform=[[1], [2]])
        assert np.all(conn.transform.init == np.array([[1], [2]]))
        assert np.all(full_transform(conn) == np.array([[0, 0], [0, 1], [0, 2]]))

        # Full slices that can be optimized away
        conn = nengo.Connection(ens3[:], ens3, transform=2)
        assert np.all(conn.transform.init == np.array(2))
        assert np.all(full_transform(conn) == np.array(2))

        # Pre slice with 1x1 transform on 2x2 slices -> 2x3 transform
        conn = nengo.Connection(neurons3[:2], ens2, transform=-1)
        assert np.all(conn.transform.init == np.array(-1))
        assert np.all(full_transform(conn) == np.array([[-1, 0, 0], [0, -1, 0]]))

        # Both slices with 1x1 transform on 2x2 slices -> 3x3 transform
        conn = nengo.Connection(neurons3[1:], neurons3[::2], transform=-1)
        assert np.all(conn.transform.init == np.array(-1))
        assert np.all(
            full_transform(conn) == np.array([[0, -1, 0], [0, 0, 0], [0, 0, -1]])
        )

        # Both slices with 2x2 transform -> 3x3 transform
        conn = nengo.Connection(node3[[0, 2]], neurons3[1:], transform=[[1, 2], [3, 4]])
        assert np.all(conn.transform.init == np.array([[1, 2], [3, 4]]))
        assert np.all(
            full_transform(conn) == np.array([[0, 0, 0], [1, 0, 2], [3, 0, 4]])
        )

        # Both slices with 2x3 transform -> 3x3 transform... IN REVERSE!
        conn = nengo.Connection(
            neurons3[::-1], neurons3[[2, 0]], transform=[[1, 2, 3], [4, 5, 6]]
        )
        assert np.all(conn.transform.init == np.array([[1, 2, 3], [4, 5, 6]]))
        assert np.all(
            full_transform(conn) == np.array([[6, 5, 4], [0, 0, 0], [3, 2, 1]])
        )

        # Both slices using lists
        conn = nengo.Connection(
            neurons3[[1, 0, 2]], neurons3[[2, 1]], transform=[[1, 2, 3], [4, 5, 6]]
        )
        assert np.all(conn.transform.init == np.array([[1, 2, 3], [4, 5, 6]]))
        assert np.all(
            full_transform(conn) == np.array([[0, 0, 0], [5, 4, 6], [2, 1, 3]])
        )

        # using vector
        conn = nengo.Connection(ens3[[1, 0, 2]], ens3[[2, 0, 1]], transform=[1, 2, 3])
        assert np.all(conn.transform.init == np.array([1, 2, 3]))
        assert np.all(
            full_transform(conn) == np.array([[2, 0, 0], [0, 0, 3], [0, 1, 0]])
        )

        # using vector 1D
        conn = nengo.Connection(ens1, ens1, transform=[5])
        assert full_transform(conn).ndim != 1
        assert np.all(full_transform(conn) == 5)

        # using vector and lists
        conn = nengo.Connection(ens3[[1, 0, 2]], ens3[[2, 0, 1]], transform=[1, 2, 3])
        assert np.all(conn.transform.init == np.array([1, 2, 3]))
        assert np.all(
            full_transform(conn) == np.array([[2, 0, 0], [0, 0, 3], [0, 1, 0]])
        )

        # using multi-index lists
        conn = nengo.Connection(ens3, ens2[[0, 1, 0]])
        assert np.all(full_transform(conn) == np.array([[1, 0, 1], [0, 1, 0]]))


def test_graphviz_moved():
    with pytest.raises(MovedError):
        generate_graphviz()


def test_remove_passthrough():
    """Test scanning through a model and removing Nodes with output=None"""

    model = nengo.Network()
    with model:
        D = 3
        input = nengo.Node([1] * D, label="input")
        a = nengo.networks.EnsembleArray(50, D, label="a")
        b = nengo.networks.EnsembleArray(50, D, label="b")

        def printout(t, x):
            logging.info("%s, %s", t, x)

        output = nengo.Node(printout, size_in=D, label="output")

        nengo.Connection(input, a.input, synapse=0.01)
        nengo.Connection(a.output, b.input, synapse=0.01)
        nengo.Connection(b.output, b.input, synapse=0.01, transform=0.9)
        nengo.Connection(a.output, a.input, synapse=0.01, transform=np.ones((D, D)))
        nengo.Connection(b.output, output, synapse=0.01)

    objs, conns = remove_passthrough_nodes(*objs_and_connections(model))

    assert len(objs) == 8
    assert len(conns) == 21


def test_remove_passthrough_bg():
    """Test scanning through a model and removing Nodes with output=None"""

    model = nengo.Network()
    with model:
        D = 3
        input = nengo.Node([1] * D, label="input")

        def printout(t, x):
            logging.info("%s, %s", t, x)

        output = nengo.Node(printout, size_in=D, label="output")
        bg = nengo.networks.BasalGanglia(D, 20)
        nengo.Connection(input, bg.input, synapse=0.01)
        nengo.Connection(bg.output, output, synapse=0.01)

    objs, conns = remove_passthrough_nodes(*objs_and_connections(model))

    assert len(objs) == 17
    assert len(conns) == 42


def test_passthrough_errors():
    """Test errors removing Nodes with output=None"""

    model = nengo.Network()
    with model:
        a = nengo.Ensemble(10, 1)
        b = nengo.Ensemble(10, 1)
        node = nengo.Node(None, size_in=1)
        nengo.Connection(a, node, synapse=0.01)
        nengo.Connection(node, b, synapse=0.01)
    with pytest.raises(Unconvertible):
        remove_passthrough_nodes(*objs_and_connections(model))

    model = nengo.Network()
    with model:
        node = nengo.Node(None, size_in=1)
        nengo.Connection(node, node, synapse=0.01)
    with pytest.raises(Unconvertible):
        remove_passthrough_nodes(*objs_and_connections(model))


@pytest.mark.filterwarnings("ignore:'Node.size_out' is being overwritten")
def test_create_replacement_connection_errors():
    with nengo.Network():
        a = nengo.Ensemble(10, 1)
        b = nengo.Ensemble(10, 1)
        v = nengo.Node(size_in=1)
        c1 = nengo.Connection(a, v, synapse=0.005)
        c2 = nengo.Connection(v, b, synapse=0.003)

        with pytest.raises(Unconvertible, match="Cannot merge two filters"):
            _create_replacement_connection(c1, c2)

        # set function on c2 (need to set v.output temporarily to avoid API validation)
        v.output = lambda t, x: x + 1
        c2.function = lambda x: x ** 2
        v.output = None
        c1.synapse = None
        with pytest.raises(
            Unconvertible, match="Cannot remove a connection with a function"
        ):
            _create_replacement_connection(c1, c2)



from nengo.connection import Connection
from nengo.ensemble import Ensemble
from nengo.exceptions import ObsoleteError
from nengo.network import Network
from nengo.node import Node


class Integrator(Network):
    """An ensemble that accumulates input and maintains state.

    This is accomplished through scaling the input signal and recurrently
    connecting an ensemble to itself to maintain state.

    Parameters
    ----------
    recurrent_tau : float
        Time constant on the recurrent connection.
    n_neurons : int
        Number of neurons in the recurrently connected ensemble.
    dimensions : int
        Dimensionality of the input signal and ensemble.
    **kwargs
        Keyword arguments passed through to ``nengo.Network``
        like 'label' and 'seed'.

    Attributes
    ----------
    ensemble : Ensemble
        The recurrently connected ensemble.
    input : Node
        Provides the input signal.
    """

    def __init__(self, recurrent_tau, n_neurons, dimensions, **kwargs):
        if "net" in kwargs:
            raise ObsoleteError("The 'net' argument is no longer supported.")
        kwargs.setdefault("label", "Integrator")
        super().__init__(**kwargs)

        with self:
            self.input = Node(size_in=dimensions)
            self.ensemble = Ensemble(n_neurons, dimensions=dimensions)
            Connection(self.ensemble, self.ensemble, synapse=recurrent_tau)
            Connection(self.input, self.ensemble, transform=recurrent_tau, synapse=None)

        self.output = self.ensemble



import numpy as np

from nengo.builder.builder import Builder
from nengo.builder.connection import slice_signal
from nengo.builder.operator import Copy, DotInc, Operator, Reset
from nengo.builder.signal import Signal
from nengo.connection import LearningRule
from nengo.ensemble import Ensemble, Neurons
from nengo.exceptions import BuildError
from nengo.learning_rules import BCM, PES, RLS, Oja, Voja
from nengo.node import Node


class SimPES(Operator):
    r"""Calculate connection weight change according to the PES rule.

    Implements the PES learning rule of the form

    .. math:: \Delta \omega_{ij} = \frac{\kappa}{n} e_j a_i

    where

    * :math:`\kappa` is a scalar learning rate,
    * :math:`n` is the number of presynaptic neurons
    * :math:`e_j` is the error for the jth output dimension, and
    * :math:`a_i` is the activity of a presynaptic neuron.

    .. versionadded:: 3.0.0

    Parameters
    ----------
    pre_filtered : Signal
        The presynaptic activity, :math:`a_i`.
    error : Signal
        The error signal, :math:`e_j`.
    delta : Signal
        The synaptic weight change to be applied, :math:`\Delta \omega_{ij}`.
    learning_rate : float
        The scalar learning rate, :math:`\kappa`.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    pre_filtered : Signal
        The presynaptic activity, :math:`a_i`.
    error : Signal
        The error signal, :math:`e_j`.
    delta : Signal
        The synaptic weight change to be applied, :math:`\Delta \omega_{ij}`.
    learning_rate : float
        The scalar learning rate, :math:`\kappa`.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[]``
    3. reads ``[pre_filtered, error]``
    4. updates ``[delta]``
    """

    def __init__(self, pre_filtered, error, delta, learning_rate, tag=None):
        super().__init__(tag=tag)

        self.learning_rate = learning_rate

        self.sets = []
        self.incs = []
        self.reads = [pre_filtered, error]
        self.updates = [delta]

    @property
    def delta(self):
        return self.updates[0]

    @property
    def error(self):
        return self.reads[1]

    @property
    def pre_filtered(self):
        return self.reads[0]

    @property
    def _descstr(self):
        return f"pre={self.pre_filtered}, error={self.error} -> {self.delta}"

    def make_step(self, signals, dt, rng):
        pre_filtered = signals[self.pre_filtered]
        error = signals[self.error]
        delta = signals[self.delta]
        n_neurons = pre_filtered.shape[0]
        alpha = -self.learning_rate * dt / n_neurons

        def step_simpes():
            np.outer(alpha * error, pre_filtered, out=delta)

        return step_simpes


class SimBCM(Operator):
    r"""Calculate connection weight change according to the BCM rule.

    Implements the Bienenstock-Cooper-Munroe learning rule of the form

    .. math:: \Delta \omega_{ij} = \kappa a_j (a_j - \theta_j) a_i

    where

    * :math:`\kappa` is a scalar learning rate,
    * :math:`a_j` is the activity of a postsynaptic neuron,
    * :math:`\theta_j` is an estimate of the average :math:`a_j`, and
    * :math:`a_i` is the activity of a presynaptic neuron.

    Parameters
    ----------
    pre_filtered : Signal
        The presynaptic activity, :math:`a_i`.
    post_filtered : Signal
        The postsynaptic activity, :math:`a_j`.
    theta : Signal
        The modification threshold, :math:`\theta_j`.
    delta : Signal
        The synaptic weight change to be applied, :math:`\Delta \omega_{ij}`.
    learning_rate : float
        The scalar learning rate, :math:`\kappa`.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    delta : Signal
        The synaptic weight change to be applied, :math:`\Delta \omega_{ij}`.
    learning_rate : float
        The scalar learning rate, :math:`\kappa`.
    post_filtered : Signal
        The postsynaptic activity, :math:`a_j`.
    pre_filtered : Signal
        The presynaptic activity, :math:`a_i`.
    tag : str or None
        A label associated with the operator, for debugging purposes.
    theta : Signal
        The modification threshold, :math:`\theta_j`.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[]``
    3. reads ``[pre_filtered, post_filtered, theta]``
    4. updates ``[delta]``
    """

    def __init__(
        self, pre_filtered, post_filtered, theta, delta, learning_rate, tag=None
    ):
        super().__init__(tag=tag)
        self.learning_rate = learning_rate

        self.sets = []
        self.incs = []
        self.reads = [pre_filtered, post_filtered, theta]
        self.updates = [delta]

    @property
    def delta(self):
        return self.updates[0]

    @property
    def pre_filtered(self):
        return self.reads[0]

    @property
    def post_filtered(self):
        return self.reads[1]

    @property
    def theta(self):
        return self.reads[2]

    @property
    def _descstr(self):
        return f"pre={self.pre_filtered}, post={self.post_filtered} -> {self.delta}"

    def make_step(self, signals, dt, rng):
        pre_filtered = signals[self.pre_filtered]
        post_filtered = signals[self.post_filtered]
        theta = signals[self.theta]
        delta = signals[self.delta]
        alpha = self.learning_rate * dt

        def step_simbcm():
            delta[...] = np.outer(
                alpha * post_filtered * (post_filtered - theta), pre_filtered
            )

        return step_simbcm


class SimOja(Operator):
    r"""Calculate connection weight change according to the Oja rule.

    Implements the Oja learning rule of the form

    .. math:: \Delta \omega_{ij} = \kappa (a_i a_j - \beta a_j^2 \omega_{ij})

    where

    * :math:`\kappa` is a scalar learning rate,
    * :math:`a_i` is the activity of a presynaptic neuron,
    * :math:`a_j` is the activity of a postsynaptic neuron,
    * :math:`\beta` is a scalar forgetting rate, and
    * :math:`\omega_{ij}` is the connection weight between the two neurons.

    Parameters
    ----------
    pre_filtered : Signal
        The presynaptic activity, :math:`a_i`.
    post_filtered : Signal
        The postsynaptic activity, :math:`a_j`.
    weights : Signal
        The connection weight matrix, :math:`\omega_{ij}`.
    delta : Signal
        The synaptic weight change to be applied, :math:`\Delta \omega_{ij}`.
    learning_rate : float
        The scalar learning rate, :math:`\kappa`.
    beta : float
        The scalar forgetting rate, :math:`\beta`.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    beta : float
        The scalar forgetting rate, :math:`\beta`.
    delta : Signal
        The synaptic weight change to be applied, :math:`\Delta \omega_{ij}`.
    learning_rate : float
        The scalar learning rate, :math:`\kappa`.
    post_filtered : Signal
        The postsynaptic activity, :math:`a_j`.
    pre_filtered : Signal
        The presynaptic activity, :math:`a_i`.
    tag : str or None
        A label associated with the operator, for debugging purposes.
    weights : Signal
        The connection weight matrix, :math:`\omega_{ij}`.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[]``
    3. reads ``[pre_filtered, post_filtered, weights]``
    4. updates ``[delta]``
    """

    def __init__(
        self, pre_filtered, post_filtered, weights, delta, learning_rate, beta, tag=None
    ):
        super().__init__(tag=tag)
        self.learning_rate = learning_rate
        self.beta = beta

        self.sets = []
        self.incs = []
        self.reads = [pre_filtered, post_filtered, weights]
        self.updates = [delta]

    @property
    def delta(self):
        return self.updates[0]

    @property
    def pre_filtered(self):
        return self.reads[0]

    @property
    def post_filtered(self):
        return self.reads[1]

    @property
    def weights(self):
        return self.reads[2]

    @property
    def _descstr(self):
        return f"pre={self.pre_filtered}, post={self.post_filtered} -> {self.delta}"

    def make_step(self, signals, dt, rng):
        weights = signals[self.weights]
        pre_filtered = signals[self.pre_filtered]
        post_filtered = signals[self.post_filtered]
        delta = signals[self.delta]
        alpha = self.learning_rate * dt
        beta = self.beta

        def step_simoja():
            # perform forgetting
            post_squared = alpha * post_filtered * post_filtered
            delta[...] = -beta * weights * post_squared[:, None]

            # perform update
            delta[...] += np.outer(alpha * post_filtered, pre_filtered)

        return step_simoja


class SimVoja(Operator):
    r"""Simulates a simplified version of Oja's rule in the vector space.

    See :doc:`examples/learning/learn-associations` for details.

    Parameters
    ----------
    pre_decoded : Signal
        Decoded activity from presynaptic ensemble, :math:`a_i`.
    post_filtered : Signal
        Filtered postsynaptic activity signal.
    scaled_encoders : Signal
        2d array of encoders, multiplied by ``scale``.
    delta : Signal
        The synaptic weight change to be applied, :math:`\Delta \omega_{ij}`.
    scale : ndarray
        The length of each encoder.
    learning_signal : Signal
        Scalar signal to be multiplied by ``learning_rate``. Expected to be
        either 0 or 1 to turn learning off or on, respectively.
    learning_rate : float
        The scalar learning rate.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Attributes
    ----------
    delta : Signal
        The synaptic weight change to be applied, :math:`\Delta \omega_{ij}`.
    learning_rate : float
        The scalar learning rate.
    learning_signal : Signal
        Scalar signal to be multiplied by ``learning_rate``. Expected to be
        either 0 or 1 to turn learning off or on, respectively.
    post_filtered : Signal
        Filtered postsynaptic activity signal.
    pre_decoded : Signal
        Decoded activity from presynaptic ensemble, :math:`a_i`.
    scale : ndarray
        The length of each encoder.
    scaled_encoders : Signal
        2d array of encoders, multiplied by ``scale``.
    tag : str or None
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[]``
    3. reads ``[pre_decoded, post_filtered, scaled_encoders, learning_signal]``
    4. updates ``[delta]``
    """

    def __init__(
        self,
        pre_decoded,
        post_filtered,
        scaled_encoders,
        delta,
        scale,
        learning_signal,
        learning_rate,
        tag=None,
    ):
        super().__init__(tag=tag)
        self.scale = scale
        self.learning_rate = learning_rate

        self.sets = []
        self.incs = []
        self.reads = [pre_decoded, post_filtered, scaled_encoders, learning_signal]
        self.updates = [delta]

    @property
    def delta(self):
        return self.updates[0]

    @property
    def learning_signal(self):
        return self.reads[3]

    @property
    def pre_decoded(self):
        return self.reads[0]

    @property
    def post_filtered(self):
        return self.reads[1]

    @property
    def scaled_encoders(self):
        return self.reads[2]

    @property
    def weights(self):
        return self.reads[2]

    @property
    def _descstr(self):
        return f"pre={self.pre_decoded}, post={self.post_filtered} -> {self.delta}"

    def make_step(self, signals, dt, rng):
        pre_decoded = signals[self.pre_decoded]
        post_filtered = signals[self.post_filtered]
        scaled_encoders = signals[self.scaled_encoders]
        delta = signals[self.delta]
        learning_signal = signals[self.learning_signal]
        alpha = self.learning_rate * dt
        scale = self.scale[:, np.newaxis]

        def step_simvoja():
            delta[...] = (
                alpha
                * learning_signal
                * (
                    scale * np.outer(post_filtered, pre_decoded)
                    - post_filtered[:, np.newaxis] * scaled_encoders
                )
            )

        return step_simvoja


class SimRLS(Operator):
    r"""Calculate connection weight change according to the RLS rule.

    Implements the Recursive Least Squares (RLS) learning rule of the form

    .. math::

       g_i &= \sum_j P_{ij}(n-1) r_i \\
       P_{ij}(n) &= P_{ij}(n-1)
                  - g_i g_j / \left( 1 + \sum_{ij} r_i P_{ij}(n-1) r_j \right) \\
       \Delta \omega_{ji} &= \frac{\kappa \delta_t}{n} e_j \sum_{k} P_{ik}(n) r_k

    where :math:`r_i` are the filtered presynaptic activities, :math:`e_j` are the
    errors, :math:`\kappa` is the learning rate, :math:`\delta_t` is the simulator
    timestep, and :math:`n` is the number of presynaptic neurons.

    Parameters
    ----------
    pre_filtered : Signal
        The filtered presynaptic activity, :math:`r_i`.
    error : Signal
        The error signal, :math:`e_j`.
    delta : Signal
        The synaptic weight change to be applied, :math:`\Delta \omega_{ji}`.
    inv_gamma : ndarray
        The inverse activity matrix :math:`P_{ij}`.
    tag : str, optional
        A label associated with the operator, for debugging purposes.

    Notes
    -----
    1. sets ``[]``
    2. incs ``[]``
    3. reads ``[pre_filtered, error]``
    4. updates ``[delta, inv_gamma]``
    """

    def __init__(self, pre_filtered, error, delta, inv_gamma, tag=None):
        super().__init__(tag=tag)

        self.sets = []
        self.incs = []
        self.reads = [pre_filtered, error]
        self.updates = [delta, inv_gamma]

    @property
    def delta(self):
        return self.updates[0]

    @property
    def inv_gamma(self):
        return self.updates[1]

    @property
    def pre_filtered(self):
        return self.reads[0]

    @property
    def error(self):
        return self.reads[1]

    @property
    def _descstr(self):
        return f"pre={self.pre_filtered} -> {self.delta}"

    def make_step(self, signals, dt, rng):
        r = signals[self.pre_filtered]
        delta = signals[self.delta]
        error = signals[self.error]
        P = signals[self.inv_gamma]
        assert r.ndim == error.ndim == 1
        assert delta.ndim == P.ndim == 2
        assert np.array_equal(P, P.T), "P must be symmetric"

        def step_simrls():
            # We want to compute:
            #   P1 = P - (P r) (r^T P) / (1 + r^T P r)
            #   delta = -error^T (P1 r)
            # Taking advantage of the fact that P is symmetric (so P r = (r^T P)^T),
            # and  P1 r = (1 / (1 + r^T P r)) P r, we have:
            Pr = P.dot(r)
            rPr1 = 1 / (1 + r.dot(Pr))
            P[...] -= Pr[:, None] * (Pr[None, :] * rPr1)
            delta[...] = error[:, None] * (-rPr1 * Pr[None, :])

        return step_simrls


def get_pre_ens(conn):
    """Get the input `.Ensemble` for connection."""
    return conn.pre_obj if isinstance(conn.pre_obj, Ensemble) else conn.pre_obj.ensemble


def get_post_ens(conn):
    """Get the output `.Ensemble` for connection."""
    return (
        conn.post_obj
        if isinstance(conn.post_obj, (Ensemble, Node))
        else conn.post_obj.ensemble
    )


def build_or_passthrough(model, obj, signal):
    """Builds the obj on signal, or returns the signal if obj is None."""
    return signal if obj is None else model.build(obj, signal)


@Builder.register(LearningRule)
def build_learning_rule(model, rule):
    """Builds a `.LearningRule` object into a model.

    A brief summary of what happens in the learning rule build process,
    in order:

    1. Create a delta signal for the weight change.
    2. Add an operator to increment the weights by delta.
    3. Call build function for the learning rule type.

    The learning rule system is designed to work with multiple learning rules
    on the same connection. If only one learning rule was to be applied to the
    connection, then we could directly modify the weights, rather than
    calculating the delta here and applying it in `.build_connection`.
    However, with multiple learning rules, we must isolate each delta signal
    in case calculating the delta depends on the weights themselves,
    making the calculation depend on the order of the learning rule
    evaluations.

    Parameters
    ----------
    model : Model
        The model to build into.
    rule : LearningRule
        The learning rule to build.

    Notes
    -----
    Sets ``model.params[rule]`` to ``None``.
    """

    conn = rule.connection

    # --- Set up delta signal
    if rule.modifies == "encoders":
        post = get_post_ens(conn)
        target = model.sig[post]["encoders"]
        tag = "encoders += delta"
    elif rule.modifies in ("decoders", "weights"):
        target = model.sig[conn]["weights"]
        tag = "weights += delta"
    else:
        raise BuildError(f"Unknown target {rule.modifies!r}")

    delta = Signal(shape=target.shape, name="Delta")

    model.add_op(Copy(delta, target, inc=True, tag=tag))
    model.sig[rule]["delta"] = delta

    model.params[rule] = None  # by default, no build-time info to return
    model.build(rule.learning_rule_type, rule)  # updates delta


@Builder.register(BCM)
def build_bcm(model, bcm, rule):
    """Builds a `.BCM` object into a model.

    Calls synapse build functions to filter the pre and post activities,
    and adds a `.SimBCM` operator to the model to calculate the delta.

    Parameters
    ----------
    model : Model
        The model to build into.
    bcm : BCM
        Learning rule type to build.
    rule : LearningRule
        The learning rule object corresponding to the neuron type.

    Notes
    -----
    Does not modify ``model.params[]`` and can therefore be called
    more than once with the same `.BCM` instance.
    """

    conn = rule.connection
    pre_activities = model.sig[get_pre_ens(conn).neurons]["out"]
    post_activities = model.sig[get_post_ens(conn).neurons]["out"]
    pre_filtered = build_or_passthrough(model, bcm.pre_synapse, pre_activities)
    post_filtered = build_or_passthrough(model, bcm.post_synapse, post_activities)
    theta = build_or_passthrough(model, bcm.theta_synapse, post_activities)

    model.add_op(
        SimBCM(
            pre_filtered,
            post_filtered,
            theta,
            model.sig[rule]["delta"],
            learning_rate=bcm.learning_rate,
        )
    )

    # expose these for probes
    model.sig[rule]["theta"] = theta
    model.sig[rule]["pre_filtered"] = pre_filtered
    model.sig[rule]["post_filtered"] = post_filtered


@Builder.register(Oja)
def build_oja(model, oja, rule):
    """Builds a `.BCM` object into a model.

    Calls synapse build functions to filter the pre and post activities,
    and adds a `.SimOja` operator to the model to calculate the delta.

    Parameters
    ----------
    model : Model
        The model to build into.
    oja : Oja
        Learning rule type to build.
    rule : LearningRule
        The learning rule object corresponding to the neuron type.

    Notes
    -----
    Does not modify ``model.params[]`` and can therefore be called
    more than once with the same `.Oja` instance.
    """

    conn = rule.connection
    pre_activities = model.sig[get_pre_ens(conn).neurons]["out"]
    post_activities = model.sig[get_post_ens(conn).neurons]["out"]
    pre_filtered = build_or_passthrough(model, oja.pre_synapse, pre_activities)
    post_filtered = build_or_passthrough(model, oja.post_synapse, post_activities)

    model.add_op(
        SimOja(
            pre_filtered,
            post_filtered,
            model.sig[conn]["weights"],
            model.sig[rule]["delta"],
            learning_rate=oja.learning_rate,
            beta=oja.beta,
        )
    )

    # expose these for probes
    model.sig[rule]["pre_filtered"] = pre_filtered
    model.sig[rule]["post_filtered"] = post_filtered


@Builder.register(Voja)
def build_voja(model, voja, rule):
    """Builds a `.Voja` object into a model.

    Calls synapse build functions to filter the post activities,
    and adds a `.SimVoja` operator to the model to calculate the delta.

    Parameters
    ----------
    model : Model
        The model to build into.
    voja : Voja
        Learning rule type to build.
    rule : LearningRule
        The learning rule object corresponding to the neuron type.

    Notes
    -----
    Does not modify ``model.params[]`` and can therefore be called
    more than once with the same `.Voja` instance.
    """

    conn = rule.connection

    # Filtered post activity
    post = conn.post_obj
    post_filtered = build_or_passthrough(
        model, voja.post_synapse, model.sig[post]["out"]
    )

    # Learning signal, defaults to 1 in case no connection is made
    # and multiplied by the learning_rate * dt
    learning = Signal(shape=rule.size_in, name="Voja:learning")
    assert rule.size_in == 1
    model.add_op(Reset(learning, value=1.0))
    model.sig[rule]["in"] = learning  # optional connection will attach here

    scaled_encoders = model.sig[post]["encoders"]
    # The gain and radius are folded into the encoders during the ensemble
    # build process, so we need to make sure that the deltas are proportional
    # to this scaling factor
    encoder_scale = model.params[post].gain / post.radius
    assert post_filtered.shape == encoder_scale.shape

    model.add_op(
        SimVoja(
            pre_decoded=model.sig[conn]["out"],
            post_filtered=post_filtered,
            scaled_encoders=scaled_encoders,
            delta=model.sig[rule]["delta"],
            scale=encoder_scale,
            learning_signal=learning,
            learning_rate=voja.learning_rate,
        )
    )

    model.sig[rule]["scaled_encoders"] = scaled_encoders
    model.sig[rule]["post_filtered"] = post_filtered


@Builder.register(PES)
def build_pes(model, pes, rule):
    """Builds a `.PES` object into a model.

    Calls synapse build functions to filter the pre activities,
    and adds a `.SimPES` operator to the model to calculate the delta.

    Parameters
    ----------
    model : Model
        The model to build into.
    pes : PES
        Learning rule type to build.
    rule : LearningRule
        The learning rule object corresponding to the neuron type.

    Notes
    -----
    Does not modify ``model.params[]`` and can therefore be called
    more than once with the same `.PES` instance.
    """

    conn = rule.connection

    # Create input error signal
    error = Signal(shape=rule.size_in, name="PES:error")
    model.add_op(Reset(error))
    model.sig[rule]["in"] = error  # error connection will attach here

    # Filter pre-synaptic activities with pre_synapse
    acts = build_or_passthrough(
        model,
        pes.pre_synapse,
        slice_signal(
            model,
            model.sig[conn.pre_obj]["out"],
            conn.pre_slice,
        )
        if isinstance(conn.pre_obj, Neurons)
        else model.sig[conn.pre_obj]["out"],
    )

    if conn._to_neurons:
        # multiply error by post encoders to get a per-neuron error
        #   i.e. local_error = dot(encoders, error)
        post = get_post_ens(conn)
        if not isinstance(conn.post_slice, slice):
            raise BuildError(
                "PES learning rule does not support advanced indexing on non-decoded "
                "connections"
            )

        encoders = model.sig[post]["encoders"]
        # slice along neuron dimension if connecting to a neuron object, otherwise
        # slice along state dimension
        encoders = (
            encoders[:, conn.post_slice]
            if isinstance(conn.post_obj, Ensemble)
            else encoders[conn.post_slice, :]
        )

        local_error = Signal(shape=(encoders.shape[0],))
        model.add_op(Reset(local_error))
        model.add_op(DotInc(encoders, error, local_error, tag="PES:encode"))
    else:
        local_error = error

    model.add_op(SimPES(acts, local_error, model.sig[rule]["delta"], pes.learning_rate))

    # expose these for probes
    model.sig[rule]["error"] = error
    model.sig[rule]["activities"] = acts


@Builder.register(RLS)
def build_rls(model, rls, rule):
    """Builds an `.RLS` (Recursive Least Squares) object into a model.

    Calls synapse build functions to filter the pre activities,
    and adds a `.SimRLS` operator to the model to calculate the delta.

    Parameters
    ----------
    model : Model
        The model to build into.
    rls : RLS
        Learning rule type to build.
    rule : LearningRule
        The learning rule object corresponding to the neuron type.

    Notes
    -----
    Does not modify ``model.params[]`` and can therefore be called
    more than once with the same `.RLS` instance.
    """
    conn = rule.connection
    pre_activities = model.sig[conn.pre_obj]["out"]

    pre_filtered = (
        pre_activities
        if rls.pre_synapse is None
        else model.build(rls.pre_synapse, pre_activities)
    )

    # Create input error signal
    error = Signal(np.zeros(rule.size_in), name="RLS:error")
    model.add_op(Reset(error))
    model.sig[rule]["in"] = error

    # Create signal for running estimate of inverse correlation matrix
    assert pre_filtered.ndim == 1
    n_neurons = pre_filtered.shape[0]
    learning_rate = rls.learning_rate * model.dt / n_neurons
    inv_gamma = Signal(np.eye(n_neurons) * learning_rate, name="RLS:inv_gamma")

    model.add_op(
        SimRLS(
            pre_filtered=pre_filtered,
            error=error,
            delta=model.sig[rule]["delta"],
            inv_gamma=inv_gamma,
        )
    )

    # expose these for probes
    model.sig[rule]["pre_filtered"] = pre_filtered
    model.sig[rule]["error"] = error
    model.sig[rule]["inv_gamma"] = inv_gamma



from nengo.base import NengoObject, NengoObjectParam, ObjView
from nengo.config import Config
from nengo.connection import Connection, LearningRule
from nengo.exceptions import ValidationError
from nengo.params import ConnectionDefault, Default, NumberParam, Parameter, StringParam
from nengo.solvers import SolverParam
from nengo.synapses import SynapseParam


class TargetParam(NengoObjectParam):
    def coerce(self, probe, target):  # pylint: disable=arguments-renamed
        obj = target.obj if isinstance(target, ObjView) else target
        if not hasattr(obj, "probeable"):
            raise ValidationError(
                f"Type '{type(obj).__name__}' is not probeable",
                attr=self.name,
                obj=probe,
            )

        # do this after; better to know that type is not Probable first
        if isinstance(obj, LearningRule):
            # TODO: this special case should be able to be removed with #1310
            return Parameter.coerce(self, probe, target)
        else:
            return super().coerce(probe, target)


class AttributeParam(StringParam):
    coerce_defaults = False

    def coerce(self, probe, attr):  # pylint: disable=arguments-renamed
        value = super().coerce(probe, attr)
        if attr not in probe.obj.probeable:
            raise ValidationError(
                f"Attribute '{attr}' is not probeable on {probe.obj}.\n"
                f"Probeable attributes: {probe.obj.probeable}",
                attr=self.name,
                obj=probe,
            )
        return value


class ProbeSolverParam(SolverParam):
    def coerce(self, conn, solver):  # pylint: disable=arguments-renamed
        if solver is ConnectionDefault:
            solver = Config.default(Connection, "solver")
        return super().coerce(conn, solver)


class Probe(NengoObject):
    """A probe is an object that collects data from the simulation.

    This is to be used in any situation where you wish to gather simulation
    data (spike data, represented values, neuron voltages, etc.) for analysis.

    Probes do not directly affect the simulation.

    All Nengo objects can be probed (except Probes themselves).
    Each object has different attributes that can be probed.
    To see what is probeable for each object, print its
    ``probeable`` attribute.

    .. testcode::

       with nengo.Network():
           ens = nengo.Ensemble(10, 1)
       print(ens.probeable)

    .. testoutput::

       ('decoded_output', 'input', 'scaled_encoders')

    Parameters
    ----------
    target : Ensemble, Neurons, Node, or Connection
        The object to probe.

    attr : str, optional
        The signal to probe. Refer to the target's ``probeable`` list for
        details. If None, the first element in the ``probeable`` list
        will be used.
    sample_every : float, optional
        Sampling period in seconds. If None, the ``dt`` of the simluation
        will be used.
    synapse : Synapse, optional
        A synaptic model to filter the probed signal.
    solver : Solver, optional
        `~nengo.solvers.Solver` to compute decoders
        for probes that require them.
    label : str, optional
        A name for the probe. Used for debugging and visualization.
    seed : int, optional
        The seed used for random number generation.

    Attributes
    ----------
    attr : str or None
        The signal that will be probed. If None, the first element of the
        target's ``probeable`` list will be used.
    sample_every : float or None
        Sampling period in seconds. If None, the ``dt`` of the simluation
        will be used.
    solver : Solver or None
        `~nengo.solvers.Solver` to compute decoders. Only used for probes
        of an ensemble's decoded output.
    synapse : Synapse or None
        A synaptic model to filter the probed signal.
    target : Ensemble, Neurons, Node, or Connection
        The object to probe.
    """

    target = TargetParam("target", nonzero_size_out=True)
    attr = AttributeParam("attr", default=None, optional=True)
    sample_every = NumberParam("sample_every", default=None, optional=True, low=1e-10)
    synapse = SynapseParam("synapse", default=None, optional=True)
    solver = ProbeSolverParam("solver", default=ConnectionDefault)

    _param_init_order = ["target"]

    def __init__(
        self,
        target,
        attr=None,
        sample_every=Default,
        synapse=Default,
        solver=Default,
        label=Default,
        seed=Default,
    ):
        super().__init__(label=label, seed=seed)
        self.target = target
        self.attr = attr if attr is not None else self.obj.probeable[0]
        self.sample_every = sample_every
        self.synapse = synapse
        self.solver = solver

    def __repr__(self):
        label_txt = "" if self.label is None else f' "{self.label}"'
        return f"<Probe{label_txt} at 0x{id(self):x} of '{self.attr}' of {self.target}>"

    def __str__(self):
        label_txt = "" if self.label is None else f' "{self.label}"'
        return f"<Probe{label_txt} of '{self.attr}' of {self.target}>"

    @property
    def obj(self):
        """(Nengo object) The underlying Nengo object target."""
        return self.target.obj if isinstance(self.target, ObjView) else self.target

    @property
    def size_in(self):
        """(int) Dimensionality of the probed signal."""
        return self.target.size_out

    @property
    def size_out(self):
        """(int) Cannot connect from probes, so always 0."""
        return 0

    @property
    def slice(self):
        """(slice) The slice associated with the Nengo object target."""
        return self.target.slice if isinstance(self.target, ObjView) else None



import pytest

import nengo.utils.neurons
from nengo.exceptions import MovedError


def test_neurons_moved():
    with pytest.raises(MovedError):
        nengo.utils.neurons.spikes2events()
    with pytest.raises(MovedError):
        nengo.utils.neurons._rates_isi_events()
    with pytest.raises(MovedError):
        nengo.utils.neurons.rates_isi()
    with pytest.raises(MovedError):
        nengo.utils.neurons.lowpass_filter()
    with pytest.raises(MovedError):
        nengo.utils.neurons.rates_kernel()
    with pytest.raises(MovedError):
        nengo.utils.neurons.settled_firingrate()



import warnings
from collections import namedtuple

import numpy as np

import nengo.utils.numpy as npext
from nengo.builder.builder import Builder
from nengo.builder.operator import Copy, DotInc, Reset
from nengo.builder.signal import Signal
from nengo.dists import Distribution, get_samples
from nengo.ensemble import Ensemble
from nengo.exceptions import BuildError, NengoWarning
from nengo.neurons import Direct
from nengo.rc import rc
from nengo.utils.builder import default_n_eval_points

built_attrs = [
    "eval_points",
    "encoders",
    "intercepts",
    "max_rates",
    "scaled_encoders",
    "gain",
    "bias",
]


class BuiltEnsemble(namedtuple("BuiltEnsemble", built_attrs)):
    """Collects the parameters generated in `.build_ensemble`.

    These are stored here because in the majority of cases the equivalent
    attribute in the original ensemble is a `.Distribution`. The attributes
    of a BuiltEnsemble are the full NumPy arrays used in the simulation.

    See the `.Ensemble` documentation for more details on each parameter.

    Parameters
    ----------
    eval_points : ndarray
        Evaluation points.
    encoders : ndarray
        Normalized encoders.
    intercepts : ndarray
        X-intercept of each neuron.
    max_rates : ndarray
        Maximum firing rates for each neuron.
    scaled_encoders : ndarray
        Normalized encoders scaled by the gain and radius.
        This quantity is used in the actual simulation, unlike ``encoders``.
    gain : ndarray
        Gain of each neuron.
    bias : ndarray
        Bias current injected into each neuron.
    """

    __slots__ = ()

    def __new__(
        cls, eval_points, encoders, intercepts, max_rates, scaled_encoders, gain, bias
    ):
        # Overridden to suppress the default __new__ docstring
        return tuple.__new__(
            cls,
            (eval_points, encoders, intercepts, max_rates, scaled_encoders, gain, bias),
        )


def gen_eval_points(ens, eval_points, rng, scale_eval_points=True, dtype=None):
    """Generate evaluation points for ensemble."""

    dtype = rc.float_dtype if dtype is None else dtype
    if isinstance(eval_points, Distribution):
        n_points = ens.n_eval_points
        if n_points is None:
            n_points = default_n_eval_points(ens.n_neurons, ens.dimensions)
        eval_points = eval_points.sample(n_points, ens.dimensions, rng)
        eval_points = eval_points.astype(dtype)
    else:
        if ens.n_eval_points is not None and eval_points.shape[0] != ens.n_eval_points:
            warnings.warn(
                "Number of eval_points doesn't match "
                "n_eval_points. Ignoring n_eval_points."
            )
        eval_points = np.array(eval_points, dtype=dtype)
        assert eval_points.ndim == 2

    if scale_eval_points:
        eval_points *= ens.radius  # scale by ensemble radius
    return eval_points


def get_activities(built_ens, ens, eval_points):
    """Get output of ensemble neurons for given evaluation points."""

    x = np.dot(eval_points, built_ens.encoders.T / ens.radius)
    return ens.neuron_type.rates(x, built_ens.gain, built_ens.bias)


def get_gain_bias(ens, rng=np.random, dtype=None):
    """Compute concrete gain and bias for ensemble."""

    dtype = rc.float_dtype if dtype is None else dtype
    if ens.gain is not None and ens.bias is not None:
        gain = get_samples(ens.gain, ens.n_neurons, rng=rng)
        bias = get_samples(ens.bias, ens.n_neurons, rng=rng)
        max_rates, intercepts = ens.neuron_type.max_rates_intercepts(gain, bias)

        if (
            ens.max_rates != Ensemble.max_rates.default
            or ens.intercepts != Ensemble.intercepts.default
        ):
            warnings.warn(
                NengoWarning(
                    f"Specifying the gains and biases for {ens} imposes a set of "
                    "maximum firing rates and intercepts. Further specifying "
                    "either max_rates or intercepts has no effect."
                )
            )

    elif ens.gain is not None or ens.bias is not None:
        # TODO: handle this instead of error
        raise NotImplementedError(
            f"gain or bias set for {ens}, but not both. "
            "Solving for one given the other is not implemented yet."
        )
    else:
        max_rates = get_samples(ens.max_rates, ens.n_neurons, rng=rng)
        intercepts = get_samples(ens.intercepts, ens.n_neurons, rng=rng)
        gain, bias = ens.neuron_type.gain_bias(max_rates, intercepts)
        if gain is not None and (not np.all(np.isfinite(gain)) or np.any(gain <= 0.0)):
            raise BuildError(
                f"The specified intercepts for {ens} lead to neurons with "
                "negative or non-finite gain. Please adjust the intercepts so "
                "that all gains are positive. For most neuron types (e.g., "
                "LIF neurons) this is achieved by reducing the maximum "
                "intercept value to below 1."
            )

    gain = gain.astype(dtype) if gain is not None else gain
    bias = bias.astype(dtype) if bias is not None else bias
    max_rates = max_rates.astype(dtype) if max_rates is not None else max_rates
    intercepts = intercepts.astype(dtype) if intercepts is not None else intercepts
    return gain, bias, max_rates, intercepts


@Builder.register(Ensemble)  # noqa: C901
def build_ensemble(model, ens):
    """Builds an `.Ensemble` object into a model.

    A brief summary of what happens in the ensemble build process, in order:

    1. Generate evaluation points and encoders.
    2. Normalize encoders to unit length.
    3. Determine bias and gain.
    4. Create neuron input signal
    5. Add operator for injecting bias.
    6. Call build function for neuron type.
    7. Scale encoders by gain and radius.
    8. Add operators for multiplying decoded input signal by encoders and
       incrementing the result in the neuron input signal.
    9. Call build function for injected noise.

    Some of these steps may be altered or omitted depending on the parameters
    of the ensemble, in particular the neuron type. For example, most steps are
    omitted for the `.Direct` neuron type.

    Parameters
    ----------
    model : Model
        The model to build into.
    ens : Ensemble
        The ensemble to build.

    Notes
    -----
    Sets ``model.params[ens]`` to a `.BuiltEnsemble` instance.
    """

    # Create random number generator
    rng = np.random.RandomState(model.seeds[ens])

    eval_points = gen_eval_points(ens, ens.eval_points, rng=rng, dtype=rc.float_dtype)

    # Set up signal
    model.sig[ens]["in"] = Signal(shape=ens.dimensions, name=f"{ens}.signal")
    model.add_op(Reset(model.sig[ens]["in"]))

    # Set up encoders
    if isinstance(ens.neuron_type, Direct):
        encoders = np.identity(ens.dimensions, dtype=rc.float_dtype)
    elif isinstance(ens.encoders, Distribution):
        encoders = get_samples(ens.encoders, ens.n_neurons, ens.dimensions, rng=rng)
        encoders = np.asarray(encoders, dtype=rc.float_dtype)
    else:
        encoders = npext.array(ens.encoders, min_dims=2, dtype=rc.float_dtype)
    if ens.normalize_encoders:
        encoders /= npext.norm(encoders, axis=1, keepdims=True)
    if np.any(np.isnan(encoders)):
        raise BuildError(
            f"NaNs detected in '{ens}' encoders. This usually means that you had "
            "zero-length encoders that were normalized, resulting in NaNs. Ensure all "
            "encoders have non-zero length, or set `normalize_encoders=False`."
        )

    # Build the neurons
    gain, bias, max_rates, intercepts = get_gain_bias(ens, rng, dtype=rc.float_dtype)

    if isinstance(ens.neuron_type, Direct):
        model.sig[ens.neurons]["in"] = Signal(
            shape=ens.dimensions, name=f"{ens}.neuron_in"
        )
        model.sig[ens.neurons]["out"] = model.sig[ens.neurons]["in"]
        model.add_op(Reset(model.sig[ens.neurons]["in"]))
    else:
        model.sig[ens.neurons]["in"] = Signal(
            shape=ens.n_neurons, name=f"{ens}.neuron_in"
        )
        model.sig[ens.neurons]["out"] = Signal(
            shape=ens.n_neurons, name=f"{ens}.neuron_out"
        )
        model.sig[ens.neurons]["bias"] = Signal(bias, name=f"{ens}.bias", readonly=True)
        model.add_op(Copy(model.sig[ens.neurons]["bias"], model.sig[ens.neurons]["in"]))
        # This adds the neuron's operator and sets other signals
        model.build(ens.neuron_type, ens.neurons)

    # Scale the encoders
    if isinstance(ens.neuron_type, Direct):
        scaled_encoders = encoders
    else:
        scaled_encoders = encoders * (gain / ens.radius)[:, np.newaxis]

    model.sig[ens]["encoders"] = Signal(
        scaled_encoders, name=f"{ens}.scaled_encoders", readonly=True
    )

    # Inject noise if specified
    if ens.noise is not None:
        model.build(ens.noise, sig_out=model.sig[ens.neurons]["in"], mode="inc")

    # Create output signal, using built Neurons
    model.add_op(
        DotInc(
            model.sig[ens]["encoders"],
            model.sig[ens]["in"],
            model.sig[ens.neurons]["in"],
            tag=f"{ens} encoding",
        )
    )

    # Output is neural output
    model.sig[ens]["out"] = model.sig[ens.neurons]["out"]

    model.params[ens] = BuiltEnsemble(
        eval_points=eval_points,
        encoders=encoders,
        intercepts=intercepts,
        max_rates=max_rates,
        scaled_encoders=scaled_encoders,
        gain=gain,
        bias=bias,
    )



import nengo
from nengo.processes import Piecewise
from nengo.utils.numpy import rms


def test_integrator(Simulator, plt, seed):
    model = nengo.Network(seed=seed)
    with model:
        inputs = {0: 0, 0.2: 1, 1: 0, 2: -2, 3: 0, 4: 1, 5: 0}
        input = nengo.Node(Piecewise(inputs))

        tau = 0.1
        T = nengo.networks.Integrator(tau, n_neurons=100, dimensions=1)
        nengo.Connection(input, T.input, synapse=tau)

        A = nengo.Ensemble(100, dimensions=1)
        nengo.Connection(A, A, synapse=tau)
        nengo.Connection(input, A, transform=tau, synapse=tau)

        input_p = nengo.Probe(input, sample_every=0.01)
        A_p = nengo.Probe(A, synapse=0.01, sample_every=0.01)
        T_p = nengo.Probe(T.ensemble, synapse=0.01, sample_every=0.01)

    with Simulator(model) as sim:
        sim.run(6.0)

    t = sim.trange(sample_every=0.01)
    plt.plot(t, sim.data[A_p], label="Manual")
    plt.plot(t, sim.data[T_p], label="Template")
    plt.plot(t, sim.data[input_p], "k", label="Input")
    plt.legend(loc="best")

    assert rms(sim.data[A_p] - sim.data[T_p]) < 0.1



import nengo
from nengo.processes import Piecewise


def test_inputgatedmemory(Simulator, allclose, plt, seed):
    to_memorize = 0.5
    start_memorizing = 0.4
    with nengo.Network(seed=seed) as net:
        test_input = nengo.Node(
            Piecewise({0.0: 0, 0.1: to_memorize, start_memorizing + 0.1: 0})
        )
        gate_input = nengo.Node(Piecewise({0.0: 0, start_memorizing: 1}))
        reset_input = nengo.Node(0)

        mem = nengo.networks.InputGatedMemory(150, 1, difference_gain=5.0)
        nengo.Connection(test_input, mem.input)
        nengo.Connection(gate_input, mem.gate)
        nengo.Connection(reset_input, mem.reset)

        mem_p = nengo.Probe(mem.output, synapse=0.01)

    with Simulator(net) as sim:
        sim.run(0.5)

    data = sim.data[mem_p]
    t = sim.trange()

    plt.title(f"gating at {start_memorizing:.1f} s")
    plt.plot(t, data, label="value in memory")
    plt.axhline(to_memorize, c="k", lw=2, label="value to remember")
    plt.axvline(start_memorizing, c="k", ls=":", label="start gating")
    plt.legend(loc="best")

    assert allclose(data[t < 0.1], 0, atol=0.05)
    assert allclose(data[(t > 0.2) & (t <= 0.4)], 0.5, atol=0.1)
    assert allclose(data[t > 0.4], 0.5, atol=0.1)
